// Copyright (c) 2020-2022 MathJax Consortium
// SPDX-License-Identifier: Apache-2.0
// Go declarative translation of MathJax 3.2.2 mathtools/MathtoolsMappings.ts, mathtools/MathtoolsConfiguration.ts.

package tex

var mjSourceMathtoolsMaps = []mjSourceMap{
	// mathtools/MathtoolsMappings.ts
	// Upstream line 47.
	{
		Name:    "mathtools-macros",
		Kind:    mjSourceCommandMap,
		Source:  "ts/input/tex/mathtools/MathtoolsMappings.ts",
		Line:    47,
		Methods: "MathtoolsMethods",
		Entries: []mjSourceEntry{
			{Name: "shoveleft", Value: mjSourceList{"HandleShove", "left"}},
			{Name: "shoveright", Value: mjSourceList{"HandleShove", "right"}},
			{Name: "xleftrightarrow", Value: mjSourceList{"xArrow", 8596, 10, 10}},
			{Name: "xLeftarrow", Value: mjSourceList{"xArrow", 8656, 12, 7}},
			{Name: "xRightarrow", Value: mjSourceList{"xArrow", 8658, 7, 12}},
			{Name: "xLeftrightarrow", Value: mjSourceList{"xArrow", 8660, 12, 12}},
			{Name: "xhookleftarrow", Value: mjSourceList{"xArrow", 8617, 10, 5}},
			{Name: "xhookrightarrow", Value: mjSourceList{"xArrow", 8618, 5, 10}},
			{Name: "xmapsto", Value: mjSourceList{"xArrow", 8614, 10, 10}},
			{Name: "xrightharpoondown", Value: mjSourceList{"xArrow", 8641, 5, 10}},
			{Name: "xleftharpoondown", Value: mjSourceList{"xArrow", 8637, 10, 5}},
			{Name: "xrightleftharpoons", Value: mjSourceList{"xArrow", 8652, 10, 10}},
			{Name: "xrightharpoonup", Value: mjSourceList{"xArrow", 8640, 5, 10}},
			{Name: "xleftharpoonup", Value: mjSourceList{"xArrow", 8636, 10, 5}},
			{Name: "xleftrightharpoons", Value: mjSourceList{"xArrow", 8651, 10, 10}},
			{Name: "mathllap", Value: mjSourceList{"MathLap", "l", false}},
			{Name: "mathrlap", Value: mjSourceList{"MathLap", "r", false}},
			{Name: "mathclap", Value: mjSourceList{"MathLap", "c", false}},
			{Name: "clap", Value: mjSourceList{"MtLap", "c"}},
			{Name: "textllap", Value: mjSourceList{"MtLap", "l"}},
			{Name: "textrlap", Value: mjSourceList{"MtLap", "r"}},
			{Name: "textclap", Value: mjSourceList{"MtLap", "c"}},
			{Name: "cramped", Value: "Cramped"},
			{Name: "crampedllap", Value: mjSourceList{"MathLap", "l", true}},
			{Name: "crampedrlap", Value: mjSourceList{"MathLap", "r", true}},
			{Name: "crampedclap", Value: mjSourceList{"MathLap", "c", true}},
			{Name: "crampedsubstack", Value: mjSourceList{"Macro", "\\begin{crampedsubarray}{c}#1\\end{crampedsubarray}", 1}},
			{Name: "mathmbox", Value: "MathMBox"},
			{Name: "mathmakebox", Value: "MathMakeBox"},
			{Name: "overbracket", Value: "UnderOverBracket"},
			{Name: "underbracket", Value: "UnderOverBracket"},
			{Name: "refeq", Value: "HandleRef"},
			{Name: "MoveEqLeft", Value: mjSourceList{"Macro", "\\hspace{#1em}&\\hspace{-#1em}", 1, "2"}},
			{Name: "Aboxed", Value: "Aboxed"},
			{Name: "ArrowBetweenLines", Value: "ArrowBetweenLines"},
			{Name: "vdotswithin", Value: "VDotsWithin"},
			{Name: "shortvdotswithin", Value: "ShortVDotsWithin"},
			{Name: "MTFlushSpaceAbove", Value: "FlushSpaceAbove"},
			{Name: "MTFlushSpaceBelow", Value: "FlushSpaceBelow"},
			{Name: "DeclarePairedDelimiter", Value: "DeclarePairedDelimiter"},
			{Name: "DeclarePairedDelimiterX", Value: "DeclarePairedDelimiterX"},
			{Name: "DeclarePairedDelimiterXPP", Value: "DeclarePairedDelimiterXPP"},
			{Name: "DeclarePairedDelimiters", Value: "DeclarePairedDelimiter"},
			{Name: "DeclarePairedDelimitersX", Value: "DeclarePairedDelimiterX"},
			{Name: "DeclarePairedDelimitersXPP", Value: "DeclarePairedDelimiterXPP"},
			{Name: "centercolon", Value: mjSourceList{"CenterColon", true, true}},
			{Name: "ordinarycolon", Value: mjSourceList{"CenterColon", false}},
			{Name: "MTThinColon", Value: mjSourceList{"CenterColon", true, true, true}},
			{Name: "coloneqq", Value: mjSourceList{"Relation", ":=", "≔"}},
			{Name: "Coloneqq", Value: mjSourceList{"Relation", "::=", "⩴"}},
			{Name: "coloneq", Value: mjSourceList{"Relation", ":-"}},
			{Name: "Coloneq", Value: mjSourceList{"Relation", "::-"}},
			{Name: "eqqcolon", Value: mjSourceList{"Relation", "=:", "≕"}},
			{Name: "Eqqcolon", Value: mjSourceList{"Relation", "=::"}},
			{Name: "eqcolon", Value: mjSourceList{"Relation", "-:", "∹"}},
			{Name: "Eqcolon", Value: mjSourceList{"Relation", "-::"}},
			{Name: "colonapprox", Value: mjSourceList{"Relation", ":\\approx"}},
			{Name: "Colonapprox", Value: mjSourceList{"Relation", "::\\approx"}},
			{Name: "colonsim", Value: mjSourceList{"Relation", ":\\sim"}},
			{Name: "Colonsim", Value: mjSourceList{"Relation", "::\\sim"}},
			{Name: "dblcolon", Value: mjSourceList{"Relation", "::", "∷"}},
			{Name: "nuparrow", Value: mjSourceList{"NArrow", "↑", ".06em"}},
			{Name: "ndownarrow", Value: mjSourceList{"NArrow", "↓", ".25em"}},
			{Name: "bigtimes", Value: mjSourceList{"Macro", "\\mathop{\\Large\\kern-.1em\\boldsymbol{\\times}\\kern-.1em}"}},
			{Name: "splitfrac", Value: mjSourceList{"SplitFrac", false}},
			{Name: "splitdfrac", Value: mjSourceList{"SplitFrac", true}},
			{Name: "xmathstrut", Value: "XMathStrut"},
			{Name: "prescript", Value: "Prescript"},
			{Name: "newtagform", Value: mjSourceList{"NewTagForm", false}},
			{Name: "renewtagform", Value: mjSourceList{"NewTagForm", true}},
			{Name: "usetagform", Value: "UseTagForm"},
			{Name: "adjustlimits", Value: mjSourceList{"MacroWithTemplate", "\\mathop{{#1}\\vphantom{{#3}}}_{{#2}\\vphantom{{#4}}}\\mathop{{#3}\\vphantom{{#1}}}_{{#4}\\vphantom{{#2}}}", 4, mjUndefined, "_", mjUndefined, "_"}},
			{Name: "mathtoolsset", Value: "SetOptions"},
		},
	},
	// mathtools/MathtoolsMappings.ts
	// Upstream line 154.
	{
		Name:    "mathtools-environments",
		Kind:    mjSourceEnvironmentMap,
		Source:  "ts/input/tex/mathtools/MathtoolsMappings.ts",
		Line:    154,
		Parser:  "ParseMethods.environment",
		Methods: "MathtoolsMethods",
		Entries: []mjSourceEntry{
			{Name: "dcases", Value: mjSourceList{"Array", nil, "\\{", "", "ll", nil, ".2em", "D"}},
			{Name: "rcases", Value: mjSourceList{"Array", nil, "", "\\}", "ll", nil, ".2em"}},
			{Name: "drcases", Value: mjSourceList{"Array", nil, "", "\\}", "ll", nil, ".2em", "D"}},
			{Name: "dcases*", Value: mjSourceList{"Cases", nil, "{", "", "D"}},
			{Name: "rcases*", Value: mjSourceList{"Cases", nil, "", "}"}},
			{Name: "drcases*", Value: mjSourceList{"Cases", nil, "", "}", "D"}},
			{Name: "cases*", Value: mjSourceList{"Cases", nil, "{", ""}},
			{Name: "matrix*", Value: mjSourceList{"MtMatrix", nil, nil, nil}},
			{Name: "pmatrix*", Value: mjSourceList{"MtMatrix", nil, "(", ")"}},
			{Name: "bmatrix*", Value: mjSourceList{"MtMatrix", nil, "[", "]"}},
			{Name: "Bmatrix*", Value: mjSourceList{"MtMatrix", nil, "\\{", "\\}"}},
			{Name: "vmatrix*", Value: mjSourceList{"MtMatrix", nil, "\\vert", "\\vert"}},
			{Name: "Vmatrix*", Value: mjSourceList{"MtMatrix", nil, "\\Vert", "\\Vert"}},
			{Name: "smallmatrix*", Value: mjSourceList{"MtSmallMatrix", nil, nil, nil}},
			{Name: "psmallmatrix", Value: mjSourceList{"MtSmallMatrix", nil, "(", ")", "c"}},
			{Name: "psmallmatrix*", Value: mjSourceList{"MtSmallMatrix", nil, "(", ")"}},
			{Name: "bsmallmatrix", Value: mjSourceList{"MtSmallMatrix", nil, "[", "]", "c"}},
			{Name: "bsmallmatrix*", Value: mjSourceList{"MtSmallMatrix", nil, "[", "]"}},
			{Name: "Bsmallmatrix", Value: mjSourceList{"MtSmallMatrix", nil, "\\{", "\\}", "c"}},
			{Name: "Bsmallmatrix*", Value: mjSourceList{"MtSmallMatrix", nil, "\\{", "\\}"}},
			{Name: "vsmallmatrix", Value: mjSourceList{"MtSmallMatrix", nil, "\\vert", "\\vert", "c"}},
			{Name: "vsmallmatrix*", Value: mjSourceList{"MtSmallMatrix", nil, "\\vert", "\\vert"}},
			{Name: "Vsmallmatrix", Value: mjSourceList{"MtSmallMatrix", nil, "\\Vert", "\\Vert", "c"}},
			{Name: "Vsmallmatrix*", Value: mjSourceList{"MtSmallMatrix", nil, "\\Vert", "\\Vert"}},
			{Name: "crampedsubarray", Value: mjSourceList{"Array", nil, nil, nil, nil, "0em", "0.1em", "S'", 1}},
			{Name: "multlined", Value: "MtMultlined"},
			{Name: "spreadlines", Value: mjSourceList{"SpreadLines", true}},
			{Name: "lgathered", Value: mjSourceList{"AmsEqnArray", nil, nil, nil, "l", nil, ".5em", "D"}},
			{Name: "rgathered", Value: mjSourceList{"AmsEqnArray", nil, nil, nil, "r", nil, ".5em", "D"}},
		},
	},
	// mathtools/MathtoolsMappings.ts
	// Upstream line 196.
	{
		Name:   "mathtools-delimiters",
		Kind:   mjSourceDelimiterMap,
		Source: "ts/input/tex/mathtools/MathtoolsMappings.ts",
		Line:   196,
		Parser: "ParseMethods.delimiter",
		Entries: []mjSourceEntry{
			{Name: "\\lparen", Value: "("},
			{Name: "\\rparen", Value: ")"},
		},
	},
	// mathtools/MathtoolsMappings.ts
	// Upstream line 204.
	{
		Name:    "mathtools-characters",
		Kind:    mjSourceCommandMap,
		Source:  "ts/input/tex/mathtools/MathtoolsMappings.ts",
		Line:    204,
		Methods: "MathtoolsMethods",
		Entries: []mjSourceEntry{
			{Name: ":", Value: mjSourceList{"CenterColon", true}},
		},
	},
	// mathtools/MathtoolsConfiguration.ts
	// Upstream line 47.
	{
		Name:     "mathtools-paired-delims",
		Kind:     mjSourceCommandMap,
		Source:   "ts/input/tex/mathtools/MathtoolsConfiguration.ts",
		Line:     47,
		Dynamic:  true,
		Priority: -5,
		Entries:  []mjSourceEntry{},
	},
}
