package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var EarthTones = d2themes.Theme{
	ID:   103,
	Name: "Earth Tones",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.WarmNeutral,

		B1: "#1E1303",
		B2: "#55452F",
		B3: "#9A876C",
		B4: "#C9B9A1",
		B5: "#E9DBCA",
		B6: "#FAF1E6",

		AA2: "#D35F0A",
		AA4: "#FABA8A",
		AA5: "#FFE0C7",

		AB4: "#FFE767",
		AB5: "#FFF2AA",
	},
}
