package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var DarkMauve = d2themes.Theme{
	ID:   200,
	Name: "Dark Mauve",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.DarkMauveNeutral,

		B1: "#CBA6f7",
		B2: "#CBA6f7",
		B3: "#6C7086",
		B4: "#585B70",
		B5: "#45475A",
		B6: "#313244",

		AA2: "#f38BA8",
		AA4: "#45475A",
		AA5: "#313244",

		AB4: "#45475A",
		AB5: "#313244",
	},
}
