package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var DarkFlagshipTerrastruct = d2themes.Theme{
	ID:   201,
	Name: "Dark Flagship Terrastruct",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.DarkNeutral,

		B1: "#F4F6FA",
		B2: "#6B8AFB",
		B3: "#3733E9",
		B4: "#070B67",
		B5: "#0B1197",
		B6: "#3733E9",

		AA2: "#8B5DEE",
		AA4: "#4918B1",
		AA5: "#7240DD",

		AB4: "#00607C",
		AB5: "#01799D",
	},
}
