package dagro

// Version identifies the Dagre behavioral source targeted by the D2 profile.
const Version = "3.1.1"

// GraphlibVersion identifies the Graphlib behavioral source used by the port.
const GraphlibVersion = "4.0.5"
