package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var FlagshipTerrastruct = d2themes.Theme{
	ID:   3,
	Name: "Flagship Terrastruct",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.CoolNeutral,

		B1: "#000E3D",
		B2: "#234CDA",
		B3: "#6B8AFB",
		B4: "#A6B8F8",
		B5: "#D2DBFD",
		B6: "#E7EAFF",

		AA2: "#5829DC",
		AA4: "#B4AEF8",
		AA5: "#E4DBFF",

		AB4: "#7FDBF8",
		AB5: "#C3F0FF",
	},
}
