// Copyright (c) 2017-2022 The MathJax Consortium
// SPDX-License-Identifier: Apache-2.0
// Go declarative translation of MathJax 3.2.2 ams/AmsMappings.ts, ams/AmsConfiguration.ts.

package tex

var mjSourceAMSMaps = []mjSourceMap{
	// ams/AmsMappings.ts
	// Upstream line 38.
	{
		Name:   "AMSmath-mathchar0mo",
		Kind:   mjSourceCharacterMap,
		Source: "ts/input/tex/ams/AmsMappings.ts",
		Line:   38,
		Parser: "ParseMethods.mathchar0mo",
		Entries: []mjSourceEntry{
			{Name: "iiiint", Value: mjSourceList{"⨌", mjSourceObject{{Name: "texClass", Value: 1}}}},
		},
	},
	// ams/AmsMappings.ts
	// Upstream line 45.
	{
		Name:    "AMSmath-operatorLetter",
		Kind:    mjSourceRegExpMap,
		Source:  "ts/input/tex/ams/AmsMappings.ts",
		Line:    45,
		Parser:  "AmsMethods.operatorLetter",
		Pattern: "[-*]",
		Flags:   "i",
		Entries: []mjSourceEntry{},
	},
	// ams/AmsMappings.ts
	// Upstream line 50.
	{
		Name:    "AMSmath-macros",
		Kind:    mjSourceCommandMap,
		Source:  "ts/input/tex/ams/AmsMappings.ts",
		Line:    50,
		Methods: "AmsMethods",
		Entries: []mjSourceEntry{
			{Name: "mathring", Value: mjSourceList{"Accent", "02DA"}},
			{Name: "nobreakspace", Value: "Tilde"},
			{Name: "negmedspace", Value: mjSourceList{"Spacer", -0.2222222222222222}},
			{Name: "negthickspace", Value: mjSourceList{"Spacer", -0.2777777777777778}},
			{Name: "idotsint", Value: mjSourceList{"MultiIntegral", "\\int\\cdots\\int"}},
			{Name: "dddot", Value: mjSourceList{"Accent", "20DB"}},
			{Name: "ddddot", Value: mjSourceList{"Accent", "20DC"}},
			{Name: "sideset", Value: "SideSet"},
			{Name: "boxed", Value: mjSourceList{"Macro", "\\fbox{$\\displaystyle{#1}$}", 1}},
			{Name: "tag", Value: "HandleTag"},
			{Name: "notag", Value: "HandleNoTag"},
			{Name: "eqref", Value: mjSourceList{"HandleRef", true}},
			{Name: "substack", Value: mjSourceList{"Macro", "\\begin{subarray}{c}#1\\end{subarray}", 1}},
			{Name: "injlim", Value: mjSourceList{"NamedOp", "inj&thinsp;lim"}},
			{Name: "projlim", Value: mjSourceList{"NamedOp", "proj&thinsp;lim"}},
			{Name: "varliminf", Value: mjSourceList{"Macro", "\\mathop{\\underline{\\mmlToken{mi}{lim}}}"}},
			{Name: "varlimsup", Value: mjSourceList{"Macro", "\\mathop{\\overline{\\mmlToken{mi}{lim}}}"}},
			{Name: "varinjlim", Value: mjSourceList{"Macro", "\\mathop{\\underrightarrow{\\mmlToken{mi}{lim}}}"}},
			{Name: "varprojlim", Value: mjSourceList{"Macro", "\\mathop{\\underleftarrow{\\mmlToken{mi}{lim}}}"}},
			{Name: "DeclareMathOperator", Value: "HandleDeclareOp"},
			{Name: "operatorname", Value: "HandleOperatorName"},
			{Name: "genfrac", Value: "Genfrac"},
			{Name: "frac", Value: mjSourceList{"Genfrac", "", "", "", ""}},
			{Name: "tfrac", Value: mjSourceList{"Genfrac", "", "", "", "1"}},
			{Name: "dfrac", Value: mjSourceList{"Genfrac", "", "", "", "0"}},
			{Name: "binom", Value: mjSourceList{"Genfrac", "(", ")", "0", ""}},
			{Name: "tbinom", Value: mjSourceList{"Genfrac", "(", ")", "0", "1"}},
			{Name: "dbinom", Value: mjSourceList{"Genfrac", "(", ")", "0", "0"}},
			{Name: "cfrac", Value: "CFrac"},
			{Name: "shoveleft", Value: mjSourceList{"HandleShove", "left"}},
			{Name: "shoveright", Value: mjSourceList{"HandleShove", "right"}},
			{Name: "xrightarrow", Value: mjSourceList{"xArrow", 8594, 5, 10}},
			{Name: "xleftarrow", Value: mjSourceList{"xArrow", 8592, 10, 5}},
		},
	},
	// ams/AmsMappings.ts
	// Upstream line 102.
	{
		Name:    "AMSmath-environment",
		Kind:    mjSourceEnvironmentMap,
		Source:  "ts/input/tex/ams/AmsMappings.ts",
		Line:    102,
		Parser:  "ParseMethods.environment",
		Methods: "AmsMethods",
		Entries: []mjSourceEntry{
			{Name: "equation*", Value: mjSourceList{"Equation", nil, false}},
			{Name: "eqnarray*", Value: mjSourceList{"EqnArray", nil, false, true, "rcl", "0em 0.278em", ".5em"}},
			{Name: "align", Value: mjSourceList{"EqnArray", nil, true, true, "rl", "0em 2em"}},
			{Name: "align*", Value: mjSourceList{"EqnArray", nil, false, true, "rl", "0em 2em"}},
			{Name: "multline", Value: mjSourceList{"Multline", nil, true}},
			{Name: "multline*", Value: mjSourceList{"Multline", nil, false}},
			{Name: "split", Value: mjSourceList{"EqnArray", nil, false, false, "rl", "0em"}},
			{Name: "gather", Value: mjSourceList{"EqnArray", nil, true, true, "c"}},
			{Name: "gather*", Value: mjSourceList{"EqnArray", nil, false, true, "c"}},
			{Name: "alignat", Value: mjSourceList{"AlignAt", nil, true, true}},
			{Name: "alignat*", Value: mjSourceList{"AlignAt", nil, false, true}},
			{Name: "alignedat", Value: mjSourceList{"AlignAt", nil, false, false}},
			{Name: "aligned", Value: mjSourceList{"AmsEqnArray", nil, nil, nil, "rl", "0em 2em", ".5em", "D"}},
			{Name: "gathered", Value: mjSourceList{"AmsEqnArray", nil, nil, nil, "c", nil, ".5em", "D"}},
			{Name: "xalignat", Value: mjSourceList{"XalignAt", nil, true, true}},
			{Name: "xalignat*", Value: mjSourceList{"XalignAt", nil, false, true}},
			{Name: "xxalignat", Value: mjSourceList{"XalignAt", nil, false, false}},
			{Name: "flalign", Value: mjSourceList{"FlalignArray", nil, true, false, true, "rlc", "auto auto fit"}},
			{Name: "flalign*", Value: mjSourceList{"FlalignArray", nil, false, false, true, "rlc", "auto auto fit"}},
			{Name: "subarray", Value: mjSourceList{"Array", nil, nil, nil, nil, "0em", "0.1em", "S", 1}},
			{Name: "smallmatrix", Value: mjSourceList{"Array", nil, nil, nil, "c", "0.333em", ".2em", "S", 1}},
			{Name: "matrix", Value: mjSourceList{"Array", nil, nil, nil, "c"}},
			{Name: "pmatrix", Value: mjSourceList{"Array", nil, "(", ")", "c"}},
			{Name: "bmatrix", Value: mjSourceList{"Array", nil, "[", "]", "c"}},
			{Name: "Bmatrix", Value: mjSourceList{"Array", nil, "\\{", "\\}", "c"}},
			{Name: "vmatrix", Value: mjSourceList{"Array", nil, "\\vert", "\\vert", "c"}},
			{Name: "Vmatrix", Value: mjSourceList{"Array", nil, "\\Vert", "\\Vert", "c"}},
			{Name: "cases", Value: mjSourceList{"Array", nil, "\\{", ".", "ll", nil, ".2em", "T"}},
		},
	},
	// ams/AmsMappings.ts
	// Upstream line 143.
	{
		Name:   "AMSmath-delimiter",
		Kind:   mjSourceDelimiterMap,
		Source: "ts/input/tex/ams/AmsMappings.ts",
		Line:   143,
		Parser: "ParseMethods.delimiter",
		Entries: []mjSourceEntry{
			{Name: "\\lvert", Value: mjSourceList{"|", mjSourceObject{{Name: "texClass", Value: 4}}}},
			{Name: "\\rvert", Value: mjSourceList{"|", mjSourceObject{{Name: "texClass", Value: 5}}}},
			{Name: "\\lVert", Value: mjSourceList{"‖", mjSourceObject{{Name: "texClass", Value: 4}}}},
			{Name: "\\rVert", Value: mjSourceList{"‖", mjSourceObject{{Name: "texClass", Value: 5}}}},
		},
	},
	// ams/AmsMappings.ts
	// Upstream line 154.
	{
		Name:   "AMSsymbols-mathchar0mi",
		Kind:   mjSourceCharacterMap,
		Source: "ts/input/tex/ams/AmsMappings.ts",
		Line:   154,
		Parser: "ParseMethods.mathchar0mi",
		Entries: []mjSourceEntry{
			{Name: "digamma", Value: "ϝ"},
			{Name: "varkappa", Value: "ϰ"},
			{Name: "varGamma", Value: mjSourceList{"Γ", mjSourceObject{{Name: "mathvariant", Value: "italic"}}}},
			{Name: "varDelta", Value: mjSourceList{"Δ", mjSourceObject{{Name: "mathvariant", Value: "italic"}}}},
			{Name: "varTheta", Value: mjSourceList{"Θ", mjSourceObject{{Name: "mathvariant", Value: "italic"}}}},
			{Name: "varLambda", Value: mjSourceList{"Λ", mjSourceObject{{Name: "mathvariant", Value: "italic"}}}},
			{Name: "varXi", Value: mjSourceList{"Ξ", mjSourceObject{{Name: "mathvariant", Value: "italic"}}}},
			{Name: "varPi", Value: mjSourceList{"Π", mjSourceObject{{Name: "mathvariant", Value: "italic"}}}},
			{Name: "varSigma", Value: mjSourceList{"Σ", mjSourceObject{{Name: "mathvariant", Value: "italic"}}}},
			{Name: "varUpsilon", Value: mjSourceList{"Υ", mjSourceObject{{Name: "mathvariant", Value: "italic"}}}},
			{Name: "varPhi", Value: mjSourceList{"Φ", mjSourceObject{{Name: "mathvariant", Value: "italic"}}}},
			{Name: "varPsi", Value: mjSourceList{"Ψ", mjSourceObject{{Name: "mathvariant", Value: "italic"}}}},
			{Name: "varOmega", Value: mjSourceList{"Ω", mjSourceObject{{Name: "mathvariant", Value: "italic"}}}},
			{Name: "beth", Value: "ℶ"},
			{Name: "gimel", Value: "ℷ"},
			{Name: "daleth", Value: "ℸ"},
			{Name: "backprime", Value: mjSourceList{"‵", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "hslash", Value: "ℏ"},
			{Name: "varnothing", Value: mjSourceList{"∅", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "blacktriangle", Value: "▴"},
			{Name: "triangledown", Value: mjSourceList{"▽", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "blacktriangledown", Value: "▾"},
			{Name: "square", Value: "◻"},
			{Name: "Box", Value: "◻"},
			{Name: "blacksquare", Value: "◼"},
			{Name: "lozenge", Value: "◊"},
			{Name: "Diamond", Value: "◊"},
			{Name: "blacklozenge", Value: "⧫"},
			{Name: "circledS", Value: mjSourceList{"Ⓢ", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "bigstar", Value: "★"},
			{Name: "sphericalangle", Value: "∢"},
			{Name: "measuredangle", Value: "∡"},
			{Name: "nexists", Value: "∄"},
			{Name: "complement", Value: "∁"},
			{Name: "mho", Value: "℧"},
			{Name: "eth", Value: mjSourceList{"ð", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "Finv", Value: "Ⅎ"},
			{Name: "diagup", Value: "╱"},
			{Name: "Game", Value: "⅁"},
			{Name: "diagdown", Value: "╲"},
			{Name: "Bbbk", Value: mjSourceList{"k", mjSourceObject{{Name: "mathvariant", Value: "double-struck"}}}},
			{Name: "yen", Value: "¥"},
			{Name: "circledR", Value: "®"},
			{Name: "checkmark", Value: "✓"},
			{Name: "maltese", Value: "✠"},
		},
	},
	// ams/AmsMappings.ts
	// Upstream line 217.
	{
		Name:   "AMSsymbols-mathchar0mo",
		Kind:   mjSourceCharacterMap,
		Source: "ts/input/tex/ams/AmsMappings.ts",
		Line:   217,
		Parser: "ParseMethods.mathchar0mo",
		Entries: []mjSourceEntry{
			{Name: "dotplus", Value: "∔"},
			{Name: "ltimes", Value: "⋉"},
			{Name: "smallsetminus", Value: mjSourceList{"∖", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "rtimes", Value: "⋊"},
			{Name: "Cap", Value: "⋒"},
			{Name: "doublecap", Value: "⋒"},
			{Name: "leftthreetimes", Value: "⋋"},
			{Name: "Cup", Value: "⋓"},
			{Name: "doublecup", Value: "⋓"},
			{Name: "rightthreetimes", Value: "⋌"},
			{Name: "barwedge", Value: "⊼"},
			{Name: "curlywedge", Value: "⋏"},
			{Name: "veebar", Value: "⊻"},
			{Name: "curlyvee", Value: "⋎"},
			{Name: "doublebarwedge", Value: "⩞"},
			{Name: "boxminus", Value: "⊟"},
			{Name: "circleddash", Value: "⊝"},
			{Name: "boxtimes", Value: "⊠"},
			{Name: "circledast", Value: "⊛"},
			{Name: "boxdot", Value: "⊡"},
			{Name: "circledcirc", Value: "⊚"},
			{Name: "boxplus", Value: "⊞"},
			{Name: "centerdot", Value: mjSourceList{"⋅", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "divideontimes", Value: "⋇"},
			{Name: "intercal", Value: "⊺"},
			{Name: "leqq", Value: "≦"},
			{Name: "geqq", Value: "≧"},
			{Name: "leqslant", Value: "⩽"},
			{Name: "geqslant", Value: "⩾"},
			{Name: "eqslantless", Value: "⪕"},
			{Name: "eqslantgtr", Value: "⪖"},
			{Name: "lesssim", Value: "≲"},
			{Name: "gtrsim", Value: "≳"},
			{Name: "lessapprox", Value: "⪅"},
			{Name: "gtrapprox", Value: "⪆"},
			{Name: "approxeq", Value: "≊"},
			{Name: "lessdot", Value: "⋖"},
			{Name: "gtrdot", Value: "⋗"},
			{Name: "lll", Value: "⋘"},
			{Name: "llless", Value: "⋘"},
			{Name: "ggg", Value: "⋙"},
			{Name: "gggtr", Value: "⋙"},
			{Name: "lessgtr", Value: "≶"},
			{Name: "gtrless", Value: "≷"},
			{Name: "lesseqgtr", Value: "⋚"},
			{Name: "gtreqless", Value: "⋛"},
			{Name: "lesseqqgtr", Value: "⪋"},
			{Name: "gtreqqless", Value: "⪌"},
			{Name: "doteqdot", Value: "≑"},
			{Name: "Doteq", Value: "≑"},
			{Name: "eqcirc", Value: "≖"},
			{Name: "risingdotseq", Value: "≓"},
			{Name: "circeq", Value: "≗"},
			{Name: "fallingdotseq", Value: "≒"},
			{Name: "triangleq", Value: "≜"},
			{Name: "backsim", Value: "∽"},
			{Name: "thicksim", Value: mjSourceList{"∼", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "backsimeq", Value: "⋍"},
			{Name: "thickapprox", Value: mjSourceList{"≈", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "subseteqq", Value: "⫅"},
			{Name: "supseteqq", Value: "⫆"},
			{Name: "Subset", Value: "⋐"},
			{Name: "Supset", Value: "⋑"},
			{Name: "sqsubset", Value: "⊏"},
			{Name: "sqsupset", Value: "⊐"},
			{Name: "preccurlyeq", Value: "≼"},
			{Name: "succcurlyeq", Value: "≽"},
			{Name: "curlyeqprec", Value: "⋞"},
			{Name: "curlyeqsucc", Value: "⋟"},
			{Name: "precsim", Value: "≾"},
			{Name: "succsim", Value: "≿"},
			{Name: "precapprox", Value: "⪷"},
			{Name: "succapprox", Value: "⪸"},
			{Name: "vartriangleleft", Value: "⊲"},
			{Name: "lhd", Value: "⊲"},
			{Name: "vartriangleright", Value: "⊳"},
			{Name: "rhd", Value: "⊳"},
			{Name: "trianglelefteq", Value: "⊴"},
			{Name: "unlhd", Value: "⊴"},
			{Name: "trianglerighteq", Value: "⊵"},
			{Name: "unrhd", Value: "⊵"},
			{Name: "vDash", Value: mjSourceList{"⊨", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "Vdash", Value: "⊩"},
			{Name: "Vvdash", Value: "⊪"},
			{Name: "smallsmile", Value: mjSourceList{"⌣", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "shortmid", Value: mjSourceList{"∣", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "smallfrown", Value: mjSourceList{"⌢", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "shortparallel", Value: mjSourceList{"∥", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "bumpeq", Value: "≏"},
			{Name: "between", Value: "≬"},
			{Name: "Bumpeq", Value: "≎"},
			{Name: "pitchfork", Value: "⋔"},
			{Name: "varpropto", Value: mjSourceList{"∝", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "backepsilon", Value: "∍"},
			{Name: "blacktriangleleft", Value: "◂"},
			{Name: "blacktriangleright", Value: "▸"},
			{Name: "therefore", Value: "∴"},
			{Name: "because", Value: "∵"},
			{Name: "eqsim", Value: "≂"},
			{Name: "vartriangle", Value: mjSourceList{"△", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "Join", Value: "⋈"},
			{Name: "nless", Value: "≮"},
			{Name: "ngtr", Value: "≯"},
			{Name: "nleq", Value: "≰"},
			{Name: "ngeq", Value: "≱"},
			{Name: "nleqslant", Value: mjSourceList{"⪇", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "ngeqslant", Value: mjSourceList{"⪈", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "nleqq", Value: mjSourceList{"≰", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "ngeqq", Value: mjSourceList{"≱", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "lneq", Value: "⪇"},
			{Name: "gneq", Value: "⪈"},
			{Name: "lneqq", Value: "≨"},
			{Name: "gneqq", Value: "≩"},
			{Name: "lvertneqq", Value: mjSourceList{"≨", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "gvertneqq", Value: mjSourceList{"≩", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "lnsim", Value: "⋦"},
			{Name: "gnsim", Value: "⋧"},
			{Name: "lnapprox", Value: "⪉"},
			{Name: "gnapprox", Value: "⪊"},
			{Name: "nprec", Value: "⊀"},
			{Name: "nsucc", Value: "⊁"},
			{Name: "npreceq", Value: mjSourceList{"⋠", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "nsucceq", Value: mjSourceList{"⋡", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "precneqq", Value: "⪵"},
			{Name: "succneqq", Value: "⪶"},
			{Name: "precnsim", Value: "⋨"},
			{Name: "succnsim", Value: "⋩"},
			{Name: "precnapprox", Value: "⪹"},
			{Name: "succnapprox", Value: "⪺"},
			{Name: "nsim", Value: "≁"},
			{Name: "ncong", Value: "≇"},
			{Name: "nshortmid", Value: mjSourceList{"∤", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "nshortparallel", Value: mjSourceList{"∦", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "nmid", Value: "∤"},
			{Name: "nparallel", Value: "∦"},
			{Name: "nvdash", Value: "⊬"},
			{Name: "nvDash", Value: "⊭"},
			{Name: "nVdash", Value: "⊮"},
			{Name: "nVDash", Value: "⊯"},
			{Name: "ntriangleleft", Value: "⋪"},
			{Name: "ntriangleright", Value: "⋫"},
			{Name: "ntrianglelefteq", Value: "⋬"},
			{Name: "ntrianglerighteq", Value: "⋭"},
			{Name: "nsubseteq", Value: "⊈"},
			{Name: "nsupseteq", Value: "⊉"},
			{Name: "nsubseteqq", Value: mjSourceList{"⊈", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "nsupseteqq", Value: mjSourceList{"⊉", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "subsetneq", Value: "⊊"},
			{Name: "supsetneq", Value: "⊋"},
			{Name: "varsubsetneq", Value: mjSourceList{"⊊", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "varsupsetneq", Value: mjSourceList{"⊋", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "subsetneqq", Value: "⫋"},
			{Name: "supsetneqq", Value: "⫌"},
			{Name: "varsubsetneqq", Value: mjSourceList{"⫋", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "varsupsetneqq", Value: mjSourceList{"⫌", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "leftleftarrows", Value: "⇇"},
			{Name: "rightrightarrows", Value: "⇉"},
			{Name: "leftrightarrows", Value: "⇆"},
			{Name: "rightleftarrows", Value: "⇄"},
			{Name: "Lleftarrow", Value: "⇚"},
			{Name: "Rrightarrow", Value: "⇛"},
			{Name: "twoheadleftarrow", Value: "↞"},
			{Name: "twoheadrightarrow", Value: "↠"},
			{Name: "leftarrowtail", Value: "↢"},
			{Name: "rightarrowtail", Value: "↣"},
			{Name: "looparrowleft", Value: "↫"},
			{Name: "looparrowright", Value: "↬"},
			{Name: "leftrightharpoons", Value: "⇋"},
			{Name: "rightleftharpoons", Value: mjSourceList{"⇌", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "curvearrowleft", Value: "↶"},
			{Name: "curvearrowright", Value: "↷"},
			{Name: "circlearrowleft", Value: "↺"},
			{Name: "circlearrowright", Value: "↻"},
			{Name: "Lsh", Value: "↰"},
			{Name: "Rsh", Value: "↱"},
			{Name: "upuparrows", Value: "⇈"},
			{Name: "downdownarrows", Value: "⇊"},
			{Name: "upharpoonleft", Value: "↿"},
			{Name: "upharpoonright", Value: "↾"},
			{Name: "downharpoonleft", Value: "⇃"},
			{Name: "restriction", Value: "↾"},
			{Name: "multimap", Value: "⊸"},
			{Name: "downharpoonright", Value: "⇂"},
			{Name: "leftrightsquigarrow", Value: "↭"},
			{Name: "rightsquigarrow", Value: "⇝"},
			{Name: "leadsto", Value: "⇝"},
			{Name: "dashrightarrow", Value: "⇢"},
			{Name: "dashleftarrow", Value: "⇠"},
			{Name: "nleftarrow", Value: "↚"},
			{Name: "nrightarrow", Value: "↛"},
			{Name: "nLeftarrow", Value: "⇍"},
			{Name: "nRightarrow", Value: "⇏"},
			{Name: "nleftrightarrow", Value: "↮"},
			{Name: "nLeftrightarrow", Value: "⇎"},
		},
	},
	// ams/AmsMappings.ts
	// Upstream line 428.
	{
		Name:   "AMSsymbols-delimiter",
		Kind:   mjSourceDelimiterMap,
		Source: "ts/input/tex/ams/AmsMappings.ts",
		Line:   428,
		Parser: "ParseMethods.delimiter",
		Entries: []mjSourceEntry{
			{Name: "\\ulcorner", Value: "⌜"},
			{Name: "\\urcorner", Value: "⌝"},
			{Name: "\\llcorner", Value: "⌞"},
			{Name: "\\lrcorner", Value: "⌟"},
		},
	},
	// ams/AmsMappings.ts
	// Upstream line 440.
	{
		Name:    "AMSsymbols-macros",
		Kind:    mjSourceCommandMap,
		Source:  "ts/input/tex/ams/AmsMappings.ts",
		Line:    440,
		Methods: "AmsMethods",
		Entries: []mjSourceEntry{
			{Name: "implies", Value: mjSourceList{"Macro", "\\;\\Longrightarrow\\;"}},
			{Name: "impliedby", Value: mjSourceList{"Macro", "\\;\\Longleftarrow\\;"}},
		},
	},
	// ams/AmsConfiguration.ts
	// Upstream line 46.
	{
		Name:     "ams-declare-ops",
		Kind:     mjSourceCommandMap,
		Source:   "ts/input/tex/ams/AmsConfiguration.ts",
		Line:     46,
		Dynamic:  true,
		Priority: -1,
		Entries:  []mjSourceEntry{},
	},
}
