// Copyright (c) 2018-2022 Omar Al-Ithawi and The MathJax Consortium
// SPDX-License-Identifier: Apache-2.0
// Go translation and modification of MathJax 3.2.2 ColorConstants.ts.

// Package color ports MathJax 3.2.2's TeX color extension.
package color

// NamedColors is the dvipsnames color table exported as COLORS by
// MathJax 3.2.2's ColorConstants.ts. Names are intentionally case-sensitive.
//
// The map is treated as read-only by this package. It is exported so the TeX
// configuration layer can expose the same constants when it registers the
// extension.
var NamedColors = map[string]string{
	"Apricot":        "#FBB982",
	"Aquamarine":     "#00B5BE",
	"Bittersweet":    "#C04F17",
	"Black":          "#221E1F",
	"Blue":           "#2D2F92",
	"BlueGreen":      "#00B3B8",
	"BlueViolet":     "#473992",
	"BrickRed":       "#B6321C",
	"Brown":          "#792500",
	"BurntOrange":    "#F7921D",
	"CadetBlue":      "#74729A",
	"CarnationPink":  "#F282B4",
	"Cerulean":       "#00A2E3",
	"CornflowerBlue": "#41B0E4",
	"Cyan":           "#00AEEF",
	"Dandelion":      "#FDBC42",
	"DarkOrchid":     "#A4538A",
	"Emerald":        "#00A99D",
	"ForestGreen":    "#009B55",
	"Fuchsia":        "#8C368C",
	"Goldenrod":      "#FFDF42",
	"Gray":           "#949698",
	"Green":          "#00A64F",
	"GreenYellow":    "#DFE674",
	"JungleGreen":    "#00A99A",
	"Lavender":       "#F49EC4",
	"LimeGreen":      "#8DC73E",
	"Magenta":        "#EC008C",
	"Mahogany":       "#A9341F",
	"Maroon":         "#AF3235",
	"Melon":          "#F89E7B",
	"MidnightBlue":   "#006795",
	"Mulberry":       "#A93C93",
	"NavyBlue":       "#006EB8",
	"OliveGreen":     "#3C8031",
	"Orange":         "#F58137",
	"OrangeRed":      "#ED135A",
	"Orchid":         "#AF72B0",
	"Peach":          "#F7965A",
	"Periwinkle":     "#7977B8",
	"PineGreen":      "#008B72",
	"Plum":           "#92268F",
	"ProcessBlue":    "#00B0F0",
	"Purple":         "#99479B",
	"RawSienna":      "#974006",
	"Red":            "#ED1B23",
	"RedOrange":      "#F26035",
	"RedViolet":      "#A1246B",
	"Rhodamine":      "#EF559F",
	"RoyalBlue":      "#0071BC",
	"RoyalPurple":    "#613F99",
	"RubineRed":      "#ED017D",
	"Salmon":         "#F69289",
	"SeaGreen":       "#3FBC9D",
	"Sepia":          "#671800",
	"SkyBlue":        "#46C5DD",
	"SpringGreen":    "#C6DC67",
	"Tan":            "#DA9D76",
	"TealBlue":       "#00AEB3",
	"Thistle":        "#D883B7",
	"Turquoise":      "#00B4CE",
	"Violet":         "#58429B",
	"VioletRed":      "#EF58A0",
	"White":          "#FFFFFF",
	"WildStrawberry": "#EE2967",
	"Yellow":         "#FFF200",
	"YellowGreen":    "#98CC70",
	"YellowOrange":   "#FAA21A",
}
