package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var OrangeCreamsicle = d2themes.Theme{
	ID:   101,
	Name: "Orange Creamsicle",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.WarmNeutral,

		B1: "#311602",
		B2: "#D35F0A",
		B3: "#F18F47",
		B4: "#FABA8A",
		B5: "#FFE0C7",
		B6: "#FFF6EF",

		AA2: "#13A477",
		AA4: "#A6E2D0",
		AA5: "#CAF2E6",

		AB4: "#FEEC8C",
		AB5: "#FFF8CF",
	},
}
