// Copyright (c) 2017-2022 The MathJax Consortium
// SPDX-License-Identifier: Apache-2.0
// Go declarative translation of MathJax 3.2.2 base/BaseMappings.ts, base/BaseConfiguration.ts.

package tex

var mjSourceBaseMaps = []mjSourceMap{
	// base/BaseMappings.ts
	// Upstream line 37.
	{
		Name:    "letter",
		Kind:    mjSourceRegExpMap,
		Source:  "ts/input/tex/base/BaseMappings.ts",
		Line:    37,
		Parser:  "ParseMethods.variable",
		Pattern: "[a-z]",
		Flags:   "i",
		Entries: []mjSourceEntry{},
	},
	// base/BaseMappings.ts
	// Upstream line 43.
	{
		Name:    "digit",
		Kind:    mjSourceRegExpMap,
		Source:  "ts/input/tex/base/BaseMappings.ts",
		Line:    43,
		Parser:  "ParseMethods.digit",
		Pattern: "[0-9.,]",
		Entries: []mjSourceEntry{},
	},
	// base/BaseMappings.ts
	// Upstream line 49.
	{
		Name:    "command",
		Kind:    mjSourceRegExpMap,
		Source:  "ts/input/tex/base/BaseMappings.ts",
		Line:    49,
		Parser:  "ParseMethods.controlSequence",
		Pattern: "^\\\\",
		Entries: []mjSourceEntry{},
	},
	// base/BaseMappings.ts
	// Upstream line 55.
	{
		Name:    "special",
		Kind:    mjSourceMacroMap,
		Source:  "ts/input/tex/base/BaseMappings.ts",
		Line:    55,
		Methods: "BaseMethods",
		Entries: []mjSourceEntry{
			{Name: "{", Value: "Open"},
			{Name: "}", Value: "Close"},
			{Name: "~", Value: "Tilde"},
			{Name: "^", Value: "Superscript"},
			{Name: "_", Value: "Subscript"},
			{Name: " ", Value: "Space"},
			{Name: "\t", Value: "Space"},
			{Name: "\r", Value: "Space"},
			{Name: "\n", Value: "Space"},
			{Name: "'", Value: "Prime"},
			{Name: "%", Value: "Comment"},
			{Name: "&", Value: "Entry"},
			{Name: "#", Value: "Hash"},
			{Name: " ", Value: "Space"},
			{Name: "’", Value: "Prime"},
		},
	},
	// base/BaseMappings.ts
	// Upstream line 81.
	{
		Name:   "mathchar0mi",
		Kind:   mjSourceCharacterMap,
		Source: "ts/input/tex/base/BaseMappings.ts",
		Line:   81,
		Parser: "ParseMethods.mathchar0mi",
		Entries: []mjSourceEntry{
			{Name: "alpha", Value: "α"},
			{Name: "beta", Value: "β"},
			{Name: "gamma", Value: "γ"},
			{Name: "delta", Value: "δ"},
			{Name: "epsilon", Value: "ϵ"},
			{Name: "zeta", Value: "ζ"},
			{Name: "eta", Value: "η"},
			{Name: "theta", Value: "θ"},
			{Name: "iota", Value: "ι"},
			{Name: "kappa", Value: "κ"},
			{Name: "lambda", Value: "λ"},
			{Name: "mu", Value: "μ"},
			{Name: "nu", Value: "ν"},
			{Name: "xi", Value: "ξ"},
			{Name: "omicron", Value: "ο"},
			{Name: "pi", Value: "π"},
			{Name: "rho", Value: "ρ"},
			{Name: "sigma", Value: "σ"},
			{Name: "tau", Value: "τ"},
			{Name: "upsilon", Value: "υ"},
			{Name: "phi", Value: "ϕ"},
			{Name: "chi", Value: "χ"},
			{Name: "psi", Value: "ψ"},
			{Name: "omega", Value: "ω"},
			{Name: "varepsilon", Value: "ε"},
			{Name: "vartheta", Value: "ϑ"},
			{Name: "varpi", Value: "ϖ"},
			{Name: "varrho", Value: "ϱ"},
			{Name: "varsigma", Value: "ς"},
			{Name: "varphi", Value: "φ"},
			{Name: "S", Value: mjSourceList{"§", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "aleph", Value: mjSourceList{"ℵ", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "hbar", Value: mjSourceList{"ℏ", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "imath", Value: "ı"},
			{Name: "jmath", Value: "ȷ"},
			{Name: "ell", Value: "ℓ"},
			{Name: "wp", Value: mjSourceList{"℘", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "Re", Value: mjSourceList{"ℜ", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "Im", Value: mjSourceList{"ℑ", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "partial", Value: mjSourceList{"∂", mjSourceObject{{Name: "mathvariant", Value: "italic"}}}},
			{Name: "infty", Value: mjSourceList{"∞", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "prime", Value: mjSourceList{"′", mjSourceObject{{Name: "variantForm", Value: true}}}},
			{Name: "emptyset", Value: mjSourceList{"∅", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "nabla", Value: mjSourceList{"∇", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "top", Value: mjSourceList{"⊤", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "bot", Value: mjSourceList{"⊥", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "angle", Value: mjSourceList{"∠", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "triangle", Value: mjSourceList{"△", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "backslash", Value: mjSourceList{"∖", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "forall", Value: mjSourceList{"∀", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "exists", Value: mjSourceList{"∃", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "neg", Value: mjSourceList{"¬", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "lnot", Value: mjSourceList{"¬", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "flat", Value: mjSourceList{"♭", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "natural", Value: mjSourceList{"♮", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "sharp", Value: mjSourceList{"♯", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "clubsuit", Value: mjSourceList{"♣", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "diamondsuit", Value: mjSourceList{"♢", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "heartsuit", Value: mjSourceList{"♡", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
			{Name: "spadesuit", Value: mjSourceList{"♠", mjSourceObject{{Name: "mathvariant", Value: "normal"}}}},
		},
	},
	// base/BaseMappings.ts
	// Upstream line 151.
	{
		Name:   "mathchar0mo",
		Kind:   mjSourceCharacterMap,
		Source: "ts/input/tex/base/BaseMappings.ts",
		Line:   151,
		Parser: "ParseMethods.mathchar0mo",
		Entries: []mjSourceEntry{
			{Name: "surd", Value: "√"},
			{Name: "coprod", Value: mjSourceList{"∐", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}}}},
			{Name: "bigvee", Value: mjSourceList{"⋁", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}}}},
			{Name: "bigwedge", Value: mjSourceList{"⋀", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}}}},
			{Name: "biguplus", Value: mjSourceList{"⨄", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}}}},
			{Name: "bigcap", Value: mjSourceList{"⋂", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}}}},
			{Name: "bigcup", Value: mjSourceList{"⋃", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}}}},
			{Name: "int", Value: mjSourceList{"∫", mjSourceObject{{Name: "texClass", Value: 1}}}},
			{Name: "intop", Value: mjSourceList{"∫", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}, {Name: "movablelimits", Value: true}}}},
			{Name: "iint", Value: mjSourceList{"∬", mjSourceObject{{Name: "texClass", Value: 1}}}},
			{Name: "iiint", Value: mjSourceList{"∭", mjSourceObject{{Name: "texClass", Value: 1}}}},
			{Name: "prod", Value: mjSourceList{"∏", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}}}},
			{Name: "sum", Value: mjSourceList{"∑", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}}}},
			{Name: "bigotimes", Value: mjSourceList{"⨂", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}}}},
			{Name: "bigoplus", Value: mjSourceList{"⨁", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}}}},
			{Name: "bigodot", Value: mjSourceList{"⨀", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}}}},
			{Name: "oint", Value: mjSourceList{"∮", mjSourceObject{{Name: "texClass", Value: 1}}}},
			{Name: "bigsqcup", Value: mjSourceList{"⨆", mjSourceObject{{Name: "texClass", Value: 1}, {Name: "movesupsub", Value: true}}}},
			{Name: "smallint", Value: mjSourceList{"∫", mjSourceObject{{Name: "largeop", Value: false}}}},
			{Name: "triangleleft", Value: "◃"},
			{Name: "triangleright", Value: "▹"},
			{Name: "bigtriangleup", Value: "△"},
			{Name: "bigtriangledown", Value: "▽"},
			{Name: "wedge", Value: "∧"},
			{Name: "land", Value: "∧"},
			{Name: "vee", Value: "∨"},
			{Name: "lor", Value: "∨"},
			{Name: "cap", Value: "∩"},
			{Name: "cup", Value: "∪"},
			{Name: "ddagger", Value: "‡"},
			{Name: "dagger", Value: "†"},
			{Name: "sqcap", Value: "⊓"},
			{Name: "sqcup", Value: "⊔"},
			{Name: "uplus", Value: "⊎"},
			{Name: "amalg", Value: "⨿"},
			{Name: "diamond", Value: "⋄"},
			{Name: "bullet", Value: "∙"},
			{Name: "wr", Value: "≀"},
			{Name: "div", Value: "÷"},
			{Name: "divsymbol", Value: "÷"},
			{Name: "odot", Value: mjSourceList{"⊙", mjSourceObject{{Name: "largeop", Value: false}}}},
			{Name: "oslash", Value: mjSourceList{"⊘", mjSourceObject{{Name: "largeop", Value: false}}}},
			{Name: "otimes", Value: mjSourceList{"⊗", mjSourceObject{{Name: "largeop", Value: false}}}},
			{Name: "ominus", Value: mjSourceList{"⊖", mjSourceObject{{Name: "largeop", Value: false}}}},
			{Name: "oplus", Value: mjSourceList{"⊕", mjSourceObject{{Name: "largeop", Value: false}}}},
			{Name: "mp", Value: "∓"},
			{Name: "pm", Value: "±"},
			{Name: "circ", Value: "∘"},
			{Name: "bigcirc", Value: "◯"},
			{Name: "setminus", Value: "∖"},
			{Name: "cdot", Value: "⋅"},
			{Name: "ast", Value: "∗"},
			{Name: "times", Value: "×"},
			{Name: "star", Value: "⋆"},
			{Name: "propto", Value: "∝"},
			{Name: "sqsubseteq", Value: "⊑"},
			{Name: "sqsupseteq", Value: "⊒"},
			{Name: "parallel", Value: "∥"},
			{Name: "mid", Value: "∣"},
			{Name: "dashv", Value: "⊣"},
			{Name: "vdash", Value: "⊢"},
			{Name: "leq", Value: "≤"},
			{Name: "le", Value: "≤"},
			{Name: "geq", Value: "≥"},
			{Name: "ge", Value: "≥"},
			{Name: "lt", Value: "<"},
			{Name: "gt", Value: ">"},
			{Name: "succ", Value: "≻"},
			{Name: "prec", Value: "≺"},
			{Name: "approx", Value: "≈"},
			{Name: "succeq", Value: "⪰"},
			{Name: "preceq", Value: "⪯"},
			{Name: "supset", Value: "⊃"},
			{Name: "subset", Value: "⊂"},
			{Name: "supseteq", Value: "⊇"},
			{Name: "subseteq", Value: "⊆"},
			{Name: "in", Value: "∈"},
			{Name: "ni", Value: "∋"},
			{Name: "notin", Value: "∉"},
			{Name: "owns", Value: "∋"},
			{Name: "gg", Value: "≫"},
			{Name: "ll", Value: "≪"},
			{Name: "sim", Value: "∼"},
			{Name: "simeq", Value: "≃"},
			{Name: "perp", Value: "⊥"},
			{Name: "equiv", Value: "≡"},
			{Name: "asymp", Value: "≍"},
			{Name: "smile", Value: "⌣"},
			{Name: "frown", Value: "⌢"},
			{Name: "ne", Value: "≠"},
			{Name: "neq", Value: "≠"},
			{Name: "cong", Value: "≅"},
			{Name: "doteq", Value: "≐"},
			{Name: "bowtie", Value: "⋈"},
			{Name: "models", Value: "⊨"},
			{Name: "notChar", Value: "⧸"},
			{Name: "Leftrightarrow", Value: "⇔"},
			{Name: "Leftarrow", Value: "⇐"},
			{Name: "Rightarrow", Value: "⇒"},
			{Name: "leftrightarrow", Value: "↔"},
			{Name: "leftarrow", Value: "←"},
			{Name: "gets", Value: "←"},
			{Name: "rightarrow", Value: "→"},
			{Name: "to", Value: mjSourceList{"→", mjSourceObject{{Name: "accent", Value: false}}}},
			{Name: "mapsto", Value: "↦"},
			{Name: "leftharpoonup", Value: "↼"},
			{Name: "leftharpoondown", Value: "↽"},
			{Name: "rightharpoonup", Value: "⇀"},
			{Name: "rightharpoondown", Value: "⇁"},
			{Name: "nearrow", Value: "↗"},
			{Name: "searrow", Value: "↘"},
			{Name: "nwarrow", Value: "↖"},
			{Name: "swarrow", Value: "↙"},
			{Name: "rightleftharpoons", Value: "⇌"},
			{Name: "hookrightarrow", Value: "↪"},
			{Name: "hookleftarrow", Value: "↩"},
			{Name: "longleftarrow", Value: "⟵"},
			{Name: "Longleftarrow", Value: "⟸"},
			{Name: "longrightarrow", Value: "⟶"},
			{Name: "Longrightarrow", Value: "⟹"},
			{Name: "Longleftrightarrow", Value: "⟺"},
			{Name: "longleftrightarrow", Value: "⟷"},
			{Name: "longmapsto", Value: "⟼"},
			{Name: "ldots", Value: "…"},
			{Name: "cdots", Value: "⋯"},
			{Name: "vdots", Value: "⋮"},
			{Name: "ddots", Value: "⋱"},
			{Name: "dotsc", Value: "…"},
			{Name: "dotsb", Value: "⋯"},
			{Name: "dotsm", Value: "⋯"},
			{Name: "dotsi", Value: "⋯"},
			{Name: "dotso", Value: "…"},
			{Name: "ldotp", Value: mjSourceList{".", mjSourceObject{{Name: "texClass", Value: 6}}}},
			{Name: "cdotp", Value: mjSourceList{"⋅", mjSourceObject{{Name: "texClass", Value: 6}}}},
			{Name: "colon", Value: mjSourceList{":", mjSourceObject{{Name: "texClass", Value: 6}}}},
		},
	},
	// base/BaseMappings.ts
	// Upstream line 321.
	{
		Name:   "mathchar7",
		Kind:   mjSourceCharacterMap,
		Source: "ts/input/tex/base/BaseMappings.ts",
		Line:   321,
		Parser: "ParseMethods.mathchar7",
		Entries: []mjSourceEntry{
			{Name: "Gamma", Value: "Γ"},
			{Name: "Delta", Value: "Δ"},
			{Name: "Theta", Value: "Θ"},
			{Name: "Lambda", Value: "Λ"},
			{Name: "Xi", Value: "Ξ"},
			{Name: "Pi", Value: "Π"},
			{Name: "Sigma", Value: "Σ"},
			{Name: "Upsilon", Value: "Υ"},
			{Name: "Phi", Value: "Φ"},
			{Name: "Psi", Value: "Ψ"},
			{Name: "Omega", Value: "Ω"},
			{Name: "_", Value: "_"},
			{Name: "#", Value: "#"},
			{Name: "$", Value: "$"},
			{Name: "%", Value: "%"},
			{Name: "&", Value: "&"},
			{Name: "And", Value: "&"},
		},
	},
	// base/BaseMappings.ts
	// Upstream line 346.
	{
		Name:   "delimiter",
		Kind:   mjSourceDelimiterMap,
		Source: "ts/input/tex/base/BaseMappings.ts",
		Line:   346,
		Parser: "ParseMethods.delimiter",
		Entries: []mjSourceEntry{
			{Name: "(", Value: "("},
			{Name: ")", Value: ")"},
			{Name: "[", Value: "["},
			{Name: "]", Value: "]"},
			{Name: "<", Value: "⟨"},
			{Name: ">", Value: "⟩"},
			{Name: "\\lt", Value: "⟨"},
			{Name: "\\gt", Value: "⟩"},
			{Name: "/", Value: "/"},
			{Name: "|", Value: mjSourceList{"|", mjSourceObject{{Name: "texClass", Value: 0}}}},
			{Name: ".", Value: ""},
			{Name: "\\\\", Value: "\\"},
			{Name: "\\lmoustache", Value: "⎰"},
			{Name: "\\rmoustache", Value: "⎱"},
			{Name: "\\lgroup", Value: "⟮"},
			{Name: "\\rgroup", Value: "⟯"},
			{Name: "\\arrowvert", Value: "⏐"},
			{Name: "\\Arrowvert", Value: "‖"},
			{Name: "\\bracevert", Value: "⎪"},
			{Name: "\\Vert", Value: mjSourceList{"‖", mjSourceObject{{Name: "texClass", Value: 0}}}},
			{Name: "\\|", Value: mjSourceList{"‖", mjSourceObject{{Name: "texClass", Value: 0}}}},
			{Name: "\\vert", Value: mjSourceList{"|", mjSourceObject{{Name: "texClass", Value: 0}}}},
			{Name: "\\uparrow", Value: "↑"},
			{Name: "\\downarrow", Value: "↓"},
			{Name: "\\updownarrow", Value: "↕"},
			{Name: "\\Uparrow", Value: "⇑"},
			{Name: "\\Downarrow", Value: "⇓"},
			{Name: "\\Updownarrow", Value: "⇕"},
			{Name: "\\backslash", Value: "\\"},
			{Name: "\\rangle", Value: "⟩"},
			{Name: "\\langle", Value: "⟨"},
			{Name: "\\rbrace", Value: "}"},
			{Name: "\\lbrace", Value: "{"},
			{Name: "\\}", Value: "}"},
			{Name: "\\{", Value: "{"},
			{Name: "\\rceil", Value: "⌉"},
			{Name: "\\lceil", Value: "⌈"},
			{Name: "\\rfloor", Value: "⌋"},
			{Name: "\\lfloor", Value: "⌊"},
			{Name: "\\lbrack", Value: "["},
			{Name: "\\rbrack", Value: "]"},
		},
	},
	// base/BaseMappings.ts
	// Upstream line 394.
	{
		Name:    "macros",
		Kind:    mjSourceCommandMap,
		Source:  "ts/input/tex/base/BaseMappings.ts",
		Line:    394,
		Methods: "BaseMethods",
		Entries: []mjSourceEntry{
			{Name: "displaystyle", Value: mjSourceList{"SetStyle", "D", true, 0}},
			{Name: "textstyle", Value: mjSourceList{"SetStyle", "T", false, 0}},
			{Name: "scriptstyle", Value: mjSourceList{"SetStyle", "S", false, 1}},
			{Name: "scriptscriptstyle", Value: mjSourceList{"SetStyle", "SS", false, 2}},
			{Name: "rm", Value: mjSourceList{"SetFont", "normal"}},
			{Name: "mit", Value: mjSourceList{"SetFont", "italic"}},
			{Name: "oldstyle", Value: mjSourceList{"SetFont", "-tex-oldstyle"}},
			{Name: "cal", Value: mjSourceList{"SetFont", "-tex-calligraphic"}},
			{Name: "it", Value: mjSourceList{"SetFont", "-tex-mathit"}},
			{Name: "bf", Value: mjSourceList{"SetFont", "bold"}},
			{Name: "bbFont", Value: mjSourceList{"SetFont", "double-struck"}},
			{Name: "scr", Value: mjSourceList{"SetFont", "script"}},
			{Name: "frak", Value: mjSourceList{"SetFont", "fraktur"}},
			{Name: "sf", Value: mjSourceList{"SetFont", "sans-serif"}},
			{Name: "tt", Value: mjSourceList{"SetFont", "monospace"}},
			{Name: "mathrm", Value: mjSourceList{"MathFont", "normal"}},
			{Name: "mathup", Value: mjSourceList{"MathFont", "normal"}},
			{Name: "mathnormal", Value: mjSourceList{"MathFont", ""}},
			{Name: "mathbf", Value: mjSourceList{"MathFont", "bold"}},
			{Name: "mathbfup", Value: mjSourceList{"MathFont", "bold"}},
			{Name: "mathit", Value: mjSourceList{"MathFont", "-tex-mathit"}},
			{Name: "mathbfit", Value: mjSourceList{"MathFont", "bold-italic"}},
			{Name: "mathbb", Value: mjSourceList{"MathFont", "double-struck"}},
			{Name: "Bbb", Value: mjSourceList{"MathFont", "double-struck"}},
			{Name: "mathfrak", Value: mjSourceList{"MathFont", "fraktur"}},
			{Name: "mathbffrak", Value: mjSourceList{"MathFont", "bold-fraktur"}},
			{Name: "mathscr", Value: mjSourceList{"MathFont", "script"}},
			{Name: "mathbfscr", Value: mjSourceList{"MathFont", "bold-script"}},
			{Name: "mathsf", Value: mjSourceList{"MathFont", "sans-serif"}},
			{Name: "mathsfup", Value: mjSourceList{"MathFont", "sans-serif"}},
			{Name: "mathbfsf", Value: mjSourceList{"MathFont", "bold-sans-serif"}},
			{Name: "mathbfsfup", Value: mjSourceList{"MathFont", "bold-sans-serif"}},
			{Name: "mathsfit", Value: mjSourceList{"MathFont", "sans-serif-italic"}},
			{Name: "mathbfsfit", Value: mjSourceList{"MathFont", "sans-serif-bold-italic"}},
			{Name: "mathtt", Value: mjSourceList{"MathFont", "monospace"}},
			{Name: "mathcal", Value: mjSourceList{"MathFont", "-tex-calligraphic"}},
			{Name: "mathbfcal", Value: mjSourceList{"MathFont", "-tex-bold-calligraphic"}},
			{Name: "symrm", Value: mjSourceList{"MathFont", "normal"}},
			{Name: "symup", Value: mjSourceList{"MathFont", "normal"}},
			{Name: "symnormal", Value: mjSourceList{"MathFont", ""}},
			{Name: "symbf", Value: mjSourceList{"MathFont", "bold"}},
			{Name: "symbfup", Value: mjSourceList{"MathFont", "bold"}},
			{Name: "symit", Value: mjSourceList{"MathFont", "italic"}},
			{Name: "symbfit", Value: mjSourceList{"MathFont", "bold-italic"}},
			{Name: "symbb", Value: mjSourceList{"MathFont", "double-struck"}},
			{Name: "symfrak", Value: mjSourceList{"MathFont", "fraktur"}},
			{Name: "symbffrak", Value: mjSourceList{"MathFont", "bold-fraktur"}},
			{Name: "symscr", Value: mjSourceList{"MathFont", "script"}},
			{Name: "symbfscr", Value: mjSourceList{"MathFont", "bold-script"}},
			{Name: "symsf", Value: mjSourceList{"MathFont", "sans-serif"}},
			{Name: "symsfup", Value: mjSourceList{"MathFont", "sans-serif"}},
			{Name: "symbfsf", Value: mjSourceList{"MathFont", "bold-sans-serif"}},
			{Name: "symbfsfup", Value: mjSourceList{"MathFont", "bold-sans-serif"}},
			{Name: "symsfit", Value: mjSourceList{"MathFont", "sans-serif-italic"}},
			{Name: "symbfsfit", Value: mjSourceList{"MathFont", "sans-serif-bold-italic"}},
			{Name: "symtt", Value: mjSourceList{"MathFont", "monospace"}},
			{Name: "symcal", Value: mjSourceList{"MathFont", "-tex-calligraphic"}},
			{Name: "symbfcal", Value: mjSourceList{"MathFont", "-tex-bold-calligraphic"}},
			{Name: "textrm", Value: mjSourceList{"HBox", nil, "normal"}},
			{Name: "textup", Value: mjSourceList{"HBox", nil, "normal"}},
			{Name: "textnormal", Value: mjSourceList{"HBox"}},
			{Name: "textit", Value: mjSourceList{"HBox", nil, "italic"}},
			{Name: "textbf", Value: mjSourceList{"HBox", nil, "bold"}},
			{Name: "textsf", Value: mjSourceList{"HBox", nil, "sans-serif"}},
			{Name: "texttt", Value: mjSourceList{"HBox", nil, "monospace"}},
			{Name: "tiny", Value: mjSourceList{"SetSize", 0.5}},
			{Name: "Tiny", Value: mjSourceList{"SetSize", 0.6}},
			{Name: "scriptsize", Value: mjSourceList{"SetSize", 0.7}},
			{Name: "small", Value: mjSourceList{"SetSize", 0.85}},
			{Name: "normalsize", Value: mjSourceList{"SetSize", 1}},
			{Name: "large", Value: mjSourceList{"SetSize", 1.2}},
			{Name: "Large", Value: mjSourceList{"SetSize", 1.44}},
			{Name: "LARGE", Value: mjSourceList{"SetSize", 1.73}},
			{Name: "huge", Value: mjSourceList{"SetSize", 2.07}},
			{Name: "Huge", Value: mjSourceList{"SetSize", 2.49}},
			{Name: "arcsin", Value: "NamedFn"},
			{Name: "arccos", Value: "NamedFn"},
			{Name: "arctan", Value: "NamedFn"},
			{Name: "arg", Value: "NamedFn"},
			{Name: "cos", Value: "NamedFn"},
			{Name: "cosh", Value: "NamedFn"},
			{Name: "cot", Value: "NamedFn"},
			{Name: "coth", Value: "NamedFn"},
			{Name: "csc", Value: "NamedFn"},
			{Name: "deg", Value: "NamedFn"},
			{Name: "det", Value: "NamedOp"},
			{Name: "dim", Value: "NamedFn"},
			{Name: "exp", Value: "NamedFn"},
			{Name: "gcd", Value: "NamedOp"},
			{Name: "hom", Value: "NamedFn"},
			{Name: "inf", Value: "NamedOp"},
			{Name: "ker", Value: "NamedFn"},
			{Name: "lg", Value: "NamedFn"},
			{Name: "lim", Value: "NamedOp"},
			{Name: "liminf", Value: mjSourceList{"NamedOp", "lim&thinsp;inf"}},
			{Name: "limsup", Value: mjSourceList{"NamedOp", "lim&thinsp;sup"}},
			{Name: "ln", Value: "NamedFn"},
			{Name: "log", Value: "NamedFn"},
			{Name: "max", Value: "NamedOp"},
			{Name: "min", Value: "NamedOp"},
			{Name: "Pr", Value: "NamedOp"},
			{Name: "sec", Value: "NamedFn"},
			{Name: "sin", Value: "NamedFn"},
			{Name: "sinh", Value: "NamedFn"},
			{Name: "sup", Value: "NamedOp"},
			{Name: "tan", Value: "NamedFn"},
			{Name: "tanh", Value: "NamedFn"},
			{Name: "limits", Value: mjSourceList{"Limits", 1}},
			{Name: "nolimits", Value: mjSourceList{"Limits", 0}},
			{Name: "overline", Value: mjSourceList{"UnderOver", "2015"}},
			{Name: "underline", Value: mjSourceList{"UnderOver", "2015"}},
			{Name: "overbrace", Value: mjSourceList{"UnderOver", "23DE", 1}},
			{Name: "underbrace", Value: mjSourceList{"UnderOver", "23DF", 1}},
			{Name: "overparen", Value: mjSourceList{"UnderOver", "23DC"}},
			{Name: "underparen", Value: mjSourceList{"UnderOver", "23DD"}},
			{Name: "overrightarrow", Value: mjSourceList{"UnderOver", "2192"}},
			{Name: "underrightarrow", Value: mjSourceList{"UnderOver", "2192"}},
			{Name: "overleftarrow", Value: mjSourceList{"UnderOver", "2190"}},
			{Name: "underleftarrow", Value: mjSourceList{"UnderOver", "2190"}},
			{Name: "overleftrightarrow", Value: mjSourceList{"UnderOver", "2194"}},
			{Name: "underleftrightarrow", Value: mjSourceList{"UnderOver", "2194"}},
			{Name: "overset", Value: "Overset"},
			{Name: "underset", Value: "Underset"},
			{Name: "overunderset", Value: "Overunderset"},
			{Name: "stackrel", Value: mjSourceList{"Macro", "\\mathrel{\\mathop{#2}\\limits^{#1}}", 2}},
			{Name: "stackbin", Value: mjSourceList{"Macro", "\\mathbin{\\mathop{#2}\\limits^{#1}}", 2}},
			{Name: "over", Value: "Over"},
			{Name: "overwithdelims", Value: "Over"},
			{Name: "atop", Value: "Over"},
			{Name: "atopwithdelims", Value: "Over"},
			{Name: "above", Value: "Over"},
			{Name: "abovewithdelims", Value: "Over"},
			{Name: "brace", Value: mjSourceList{"Over", "{", "}"}},
			{Name: "brack", Value: mjSourceList{"Over", "[", "]"}},
			{Name: "choose", Value: mjSourceList{"Over", "(", ")"}},
			{Name: "frac", Value: "Frac"},
			{Name: "sqrt", Value: "Sqrt"},
			{Name: "root", Value: "Root"},
			{Name: "uproot", Value: mjSourceList{"MoveRoot", "upRoot"}},
			{Name: "leftroot", Value: mjSourceList{"MoveRoot", "leftRoot"}},
			{Name: "left", Value: "LeftRight"},
			{Name: "right", Value: "LeftRight"},
			{Name: "middle", Value: "LeftRight"},
			{Name: "llap", Value: "Lap"},
			{Name: "rlap", Value: "Lap"},
			{Name: "raise", Value: "RaiseLower"},
			{Name: "lower", Value: "RaiseLower"},
			{Name: "moveleft", Value: "MoveLeftRight"},
			{Name: "moveright", Value: "MoveLeftRight"},
			{Name: ",", Value: mjSourceList{"Spacer", 0.16666666666666666}},
			{Name: ":", Value: mjSourceList{"Spacer", 0.2222222222222222}},
			{Name: ">", Value: mjSourceList{"Spacer", 0.2222222222222222}},
			{Name: ";", Value: mjSourceList{"Spacer", 0.2777777777777778}},
			{Name: "!", Value: mjSourceList{"Spacer", -0.16666666666666666}},
			{Name: "enspace", Value: mjSourceList{"Spacer", 0.5}},
			{Name: "quad", Value: mjSourceList{"Spacer", 1}},
			{Name: "qquad", Value: mjSourceList{"Spacer", 2}},
			{Name: "thinspace", Value: mjSourceList{"Spacer", 0.16666666666666666}},
			{Name: "negthinspace", Value: mjSourceList{"Spacer", -0.16666666666666666}},
			{Name: "hskip", Value: "Hskip"},
			{Name: "hspace", Value: "Hskip"},
			{Name: "kern", Value: "Hskip"},
			{Name: "mskip", Value: "Hskip"},
			{Name: "mspace", Value: "Hskip"},
			{Name: "mkern", Value: "Hskip"},
			{Name: "rule", Value: "rule"},
			{Name: "Rule", Value: mjSourceList{"Rule"}},
			{Name: "Space", Value: mjSourceList{"Rule", "blank"}},
			{Name: "nonscript", Value: "Nonscript"},
			{Name: "big", Value: mjSourceList{"MakeBig", 0, 0.85}},
			{Name: "Big", Value: mjSourceList{"MakeBig", 0, 1.15}},
			{Name: "bigg", Value: mjSourceList{"MakeBig", 0, 1.45}},
			{Name: "Bigg", Value: mjSourceList{"MakeBig", 0, 1.75}},
			{Name: "bigl", Value: mjSourceList{"MakeBig", 4, 0.85}},
			{Name: "Bigl", Value: mjSourceList{"MakeBig", 4, 1.15}},
			{Name: "biggl", Value: mjSourceList{"MakeBig", 4, 1.45}},
			{Name: "Biggl", Value: mjSourceList{"MakeBig", 4, 1.75}},
			{Name: "bigr", Value: mjSourceList{"MakeBig", 5, 0.85}},
			{Name: "Bigr", Value: mjSourceList{"MakeBig", 5, 1.15}},
			{Name: "biggr", Value: mjSourceList{"MakeBig", 5, 1.45}},
			{Name: "Biggr", Value: mjSourceList{"MakeBig", 5, 1.75}},
			{Name: "bigm", Value: mjSourceList{"MakeBig", 3, 0.85}},
			{Name: "Bigm", Value: mjSourceList{"MakeBig", 3, 1.15}},
			{Name: "biggm", Value: mjSourceList{"MakeBig", 3, 1.45}},
			{Name: "Biggm", Value: mjSourceList{"MakeBig", 3, 1.75}},
			{Name: "mathord", Value: mjSourceList{"TeXAtom", 0}},
			{Name: "mathop", Value: mjSourceList{"TeXAtom", 1}},
			{Name: "mathopen", Value: mjSourceList{"TeXAtom", 4}},
			{Name: "mathclose", Value: mjSourceList{"TeXAtom", 5}},
			{Name: "mathbin", Value: mjSourceList{"TeXAtom", 2}},
			{Name: "mathrel", Value: mjSourceList{"TeXAtom", 3}},
			{Name: "mathpunct", Value: mjSourceList{"TeXAtom", 6}},
			{Name: "mathinner", Value: mjSourceList{"TeXAtom", 7}},
			{Name: "vcenter", Value: mjSourceList{"TeXAtom", 8}},
			{Name: "buildrel", Value: "BuildRel"},
			{Name: "hbox", Value: mjSourceList{"HBox", 0}},
			{Name: "text", Value: "HBox"},
			{Name: "mbox", Value: mjSourceList{"HBox", 0}},
			{Name: "fbox", Value: "FBox"},
			{Name: "boxed", Value: mjSourceList{"Macro", "\\fbox{$\\displaystyle{#1}$}", 1}},
			{Name: "framebox", Value: "FrameBox"},
			{Name: "strut", Value: "Strut"},
			{Name: "mathstrut", Value: mjSourceList{"Macro", "\\vphantom{(}"}},
			{Name: "phantom", Value: "Phantom"},
			{Name: "vphantom", Value: mjSourceList{"Phantom", 1, 0}},
			{Name: "hphantom", Value: mjSourceList{"Phantom", 0, 1}},
			{Name: "smash", Value: "Smash"},
			{Name: "acute", Value: mjSourceList{"Accent", "00B4"}},
			{Name: "grave", Value: mjSourceList{"Accent", "0060"}},
			{Name: "ddot", Value: mjSourceList{"Accent", "00A8"}},
			{Name: "tilde", Value: mjSourceList{"Accent", "007E"}},
			{Name: "bar", Value: mjSourceList{"Accent", "00AF"}},
			{Name: "breve", Value: mjSourceList{"Accent", "02D8"}},
			{Name: "check", Value: mjSourceList{"Accent", "02C7"}},
			{Name: "hat", Value: mjSourceList{"Accent", "005E"}},
			{Name: "vec", Value: mjSourceList{"Accent", "2192"}},
			{Name: "dot", Value: mjSourceList{"Accent", "02D9"}},
			{Name: "widetilde", Value: mjSourceList{"Accent", "007E", 1}},
			{Name: "widehat", Value: mjSourceList{"Accent", "005E", 1}},
			{Name: "matrix", Value: "Matrix"},
			{Name: "array", Value: "Matrix"},
			{Name: "pmatrix", Value: mjSourceList{"Matrix", "(", ")"}},
			{Name: "cases", Value: mjSourceList{"Matrix", "{", "", "left left", nil, ".1em", nil, true}},
			{Name: "eqalign", Value: mjSourceList{"Matrix", nil, nil, "right left", "0.278em", ".5em", "D"}},
			{Name: "displaylines", Value: mjSourceList{"Matrix", nil, nil, "center", nil, ".5em", "D"}},
			{Name: "cr", Value: "Cr"},
			{Name: "\\", Value: "CrLaTeX"},
			{Name: "newline", Value: mjSourceList{"CrLaTeX", true}},
			{Name: "hline", Value: mjSourceList{"HLine", "solid"}},
			{Name: "hdashline", Value: mjSourceList{"HLine", "dashed"}},
			{Name: "eqalignno", Value: mjSourceList{"Matrix", nil, nil, "right left", "0.278em", ".5em", "D", nil, "right"}},
			{Name: "leqalignno", Value: mjSourceList{"Matrix", nil, nil, "right left", "0.278em", ".5em", "D", nil, "left"}},
			{Name: "hfill", Value: "HFill"},
			{Name: "hfil", Value: "HFill"},
			{Name: "hfilll", Value: "HFill"},
			{Name: "bmod", Value: mjSourceList{"Macro", "\\mmlToken{mo}[lspace=\"thickmathspace\" rspace=\"thickmathspace\"]{mod}"}},
			{Name: "pmod", Value: mjSourceList{"Macro", "\\pod{\\mmlToken{mi}{mod}\\kern 6mu #1}", 1}},
			{Name: "mod", Value: mjSourceList{"Macro", "\\mathchoice{\\kern18mu}{\\kern12mu}{\\kern12mu}{\\kern12mu}\\mmlToken{mi}{mod}\\,\\,#1", 1}},
			{Name: "pod", Value: mjSourceList{"Macro", "\\mathchoice{\\kern18mu}{\\kern8mu}{\\kern8mu}{\\kern8mu}(#1)", 1}},
			{Name: "iff", Value: mjSourceList{"Macro", "\\;\\Longleftrightarrow\\;"}},
			{Name: "skew", Value: mjSourceList{"Macro", "{{#2{#3\\mkern#1mu}\\mkern-#1mu}{}}", 3}},
			{Name: "pmb", Value: mjSourceList{"Macro", "\\rlap{#1}\\kern1px{#1}", 1}},
			{Name: "TeX", Value: mjSourceList{"Macro", "T\\kern-.14em\\lower.5ex{E}\\kern-.115em X"}},
			{Name: "LaTeX", Value: mjSourceList{"Macro", "L\\kern-.325em\\raise.21em{\\scriptstyle{A}}\\kern-.17em\\TeX"}},
			{Name: " ", Value: mjSourceList{"Macro", "\\text{ }"}},
			{Name: "not", Value: "Not"},
			{Name: "dots", Value: "Dots"},
			{Name: "space", Value: "Tilde"},
			{Name: " ", Value: "Tilde"},
			{Name: "begin", Value: "BeginEnd"},
			{Name: "end", Value: "BeginEnd"},
			{Name: "label", Value: "HandleLabel"},
			{Name: "ref", Value: "HandleRef"},
			{Name: "nonumber", Value: "HandleNoTag"},
			{Name: "mathchoice", Value: "MathChoice"},
			{Name: "mmlToken", Value: "MmlToken"},
		},
	},
	// base/BaseMappings.ts
	// Upstream line 703.
	{
		Name:    "environment",
		Kind:    mjSourceEnvironmentMap,
		Source:  "ts/input/tex/base/BaseMappings.ts",
		Line:    703,
		Parser:  "ParseMethods.environment",
		Methods: "BaseMethods",
		Entries: []mjSourceEntry{
			{Name: "array", Value: mjSourceList{"AlignedArray"}},
			{Name: "equation", Value: mjSourceList{"Equation", nil, true}},
			{Name: "eqnarray", Value: mjSourceList{"EqnArray", nil, true, true, "rcl", "0em 0.278em", ".5em"}},
		},
	},
	// base/BaseMappings.ts
	// Upstream line 714.
	{
		Name:   "not_remap",
		Kind:   mjSourceCharacterMap,
		Source: "ts/input/tex/base/BaseMappings.ts",
		Line:   714,
		Entries: []mjSourceEntry{
			{Name: "←", Value: "↚"},
			{Name: "→", Value: "↛"},
			{Name: "↔", Value: "↮"},
			{Name: "⇐", Value: "⇍"},
			{Name: "⇒", Value: "⇏"},
			{Name: "⇔", Value: "⇎"},
			{Name: "∈", Value: "∉"},
			{Name: "∋", Value: "∌"},
			{Name: "∣", Value: "∤"},
			{Name: "∥", Value: "∦"},
			{Name: "∼", Value: "≁"},
			{Name: "~", Value: "≁"},
			{Name: "≃", Value: "≄"},
			{Name: "≅", Value: "≇"},
			{Name: "≈", Value: "≉"},
			{Name: "≍", Value: "≭"},
			{Name: "=", Value: "≠"},
			{Name: "≡", Value: "≢"},
			{Name: "<", Value: "≮"},
			{Name: ">", Value: "≯"},
			{Name: "≤", Value: "≰"},
			{Name: "≥", Value: "≱"},
			{Name: "≲", Value: "≴"},
			{Name: "≳", Value: "≵"},
			{Name: "≶", Value: "≸"},
			{Name: "≷", Value: "≹"},
			{Name: "≺", Value: "⊀"},
			{Name: "≻", Value: "⊁"},
			{Name: "⊂", Value: "⊄"},
			{Name: "⊃", Value: "⊅"},
			{Name: "⊆", Value: "⊈"},
			{Name: "⊇", Value: "⊉"},
			{Name: "⊢", Value: "⊬"},
			{Name: "⊨", Value: "⊭"},
			{Name: "⊩", Value: "⊮"},
			{Name: "⊫", Value: "⊯"},
			{Name: "≼", Value: "⋠"},
			{Name: "≽", Value: "⋡"},
			{Name: "⊑", Value: "⋢"},
			{Name: "⊒", Value: "⋣"},
			{Name: "⊲", Value: "⋪"},
			{Name: "⊳", Value: "⋫"},
			{Name: "⊴", Value: "⋬"},
			{Name: "⊵", Value: "⋭"},
			{Name: "∃", Value: "∄"},
		},
	},
	// base/BaseConfiguration.ts
	// Upstream line 41.
	{
		Name:   "remap",
		Kind:   mjSourceCharacterMap,
		Source: "ts/input/tex/base/BaseConfiguration.ts",
		Line:   41,
		Entries: []mjSourceEntry{
			{Name: "-", Value: "−"},
			{Name: "*", Value: "∗"},
			{Name: "`", Value: "‘"},
		},
	},
}
