package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var ColorblindClear = d2themes.Theme{
	ID:   8,
	Name: "Colorblind Clear",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.CoolNeutral,

		B1: "#010E31",
		B2: "#173688",
		B3: "#5679D4",
		B4: "#84A1EC",
		B5: "#C8D6F9",
		B6: "#E5EDFF",

		AA2: "#048E63",
		AA4: "#A6E2D0",
		AA5: "#CAF2E6",

		AB4: "#FFDA90",
		AB5: "#FFF0D1",
	},
}
