// Copyright (c) 2018-2022 The MathJax Consortium
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Derived mechanically from MathJax 3.2.2
// ts/output/{common,svg}/fonts/tex.
// Code generated by generate.py; DO NOT EDIT.

package font

var delimiterTable = map[rune]Delimiter{
	0x28:   Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Stretch: []rune{0x239B, 0x239C, 0x239D}, HDW: [3]float64{0.85, 0.349, 0.875}, HasHDW: true},
	0x29:   Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Stretch: []rune{0x239E, 0x239F, 0x23A0}, HDW: [3]float64{0.85, 0.349, 0.875}, HasHDW: true},
	0x2D:   Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.778}, Stretch: []rune{0x0, 0x2212}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x2212, HasAlias: true},
	0x2F:   Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Alias: 0x2F, HasAlias: true},
	0x3D:   Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.778}, Stretch: []rune{0x0, 0x3D}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true},
	0x5B:   Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Stretch: []rune{0x23A1, 0x23A2, 0x23A3}, HDW: [3]float64{0.85, 0.349, 0.667}, HasHDW: true},
	0x5C:   Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}},
	0x5D:   Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Stretch: []rune{0x23A4, 0x23A5, 0x23A6}, HDW: [3]float64{0.85, 0.349, 0.667}, HasHDW: true},
	0x5E:   Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5, 0.556, 1, 1.444, 1.889}, Alias: 0x2C6, HasAlias: true},
	0x5F:   Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5}, Stretch: []rune{0x0, 0x2013}, HDW: [3]float64{0.285, -0.248, 0.5}, HasHDW: true, Alias: 0x2013, HasAlias: true},
	0x7B:   Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Stretch: []rune{0x23A7, 0x23AA, 0x23A9, 0x23A8}, HDW: [3]float64{0.85, 0.349, 0.889}, HasHDW: true},
	0x7C:   Delimiter{Direction: DirectionVertical, Sizes: []float64{1}, Stretch: []rune{0x0, 0x2223}, HDW: [3]float64{0.75, 0.25, 0.333}, HasHDW: true},
	0x7D:   Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Stretch: []rune{0x23AB, 0x23AA, 0x23AD, 0x23AC}, HDW: [3]float64{0.85, 0.349, 0.889}, HasHDW: true},
	0x7E:   Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5, 0.556, 1, 1.444, 1.889}, Alias: 0x2DC, HasAlias: true},
	0xAF:   Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5}, Stretch: []rune{0x0, 0xAF}, HDW: [3]float64{0.59, -0.544, 0.5}, HasHDW: true, Alias: 0xAF, HasAlias: true},
	0x2C6:  Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5, 0.556, 1, 1.444, 1.889}, Alias: 0x2C6, HasAlias: true},
	0x2C9:  Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5}, Stretch: []rune{0x0, 0xAF}, HDW: [3]float64{0.59, -0.544, 0.5}, HasHDW: true, Alias: 0xAF, HasAlias: true},
	0x2DC:  Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5, 0.556, 1, 1.444, 1.889}, Alias: 0x2DC, HasAlias: true},
	0x302:  Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5, 0.556, 1, 1.444, 1.889}, Alias: 0x2C6, HasAlias: true},
	0x303:  Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5, 0.556, 1, 1.444, 1.889}, Alias: 0x2DC, HasAlias: true},
	0x332:  Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5}, Stretch: []rune{0x0, 0x2013}, HDW: [3]float64{0.285, -0.248, 0.5}, HasHDW: true, Alias: 0x2013, HasAlias: true},
	0x2013: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5}, Stretch: []rune{0x0, 0x2013}, HDW: [3]float64{0.285, -0.248, 0.5}, HasHDW: true, Alias: 0x2013, HasAlias: true},
	0x2014: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5}, Stretch: []rune{0x0, 0x2013}, HDW: [3]float64{0.285, -0.248, 0.5}, HasHDW: true, Alias: 0x2013, HasAlias: true},
	0x2015: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5}, Stretch: []rune{0x0, 0x2013}, HDW: [3]float64{0.285, -0.248, 0.5}, HasHDW: true, Alias: 0x2013, HasAlias: true},
	0x2016: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.602, 1}, SizeVariants: []int{1, 0}, SizeChars: []rune{0x0, 0x2225}, Stretch: []rune{0x0, 0x2225}, HDW: [3]float64{0.602, 0, 0.556}, HasHDW: true},
	0x2017: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5}, Stretch: []rune{0x0, 0x2013}, HDW: [3]float64{0.285, -0.248, 0.5}, HasHDW: true, Alias: 0x2013, HasAlias: true},
	0x203E: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5}, Stretch: []rune{0x0, 0xAF}, HDW: [3]float64{0.59, -0.544, 0.5}, HasHDW: true, Alias: 0xAF, HasAlias: true},
	0x20D7: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x0, 0x2212, 0x2192}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x2192, HasAlias: true},
	0x2190: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x2190, 0x2212}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x2190, HasAlias: true},
	0x2191: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.888}, Stretch: []rune{0x2191, 0x23D0}, HDW: [3]float64{0.6, 0, 0.667}, HasHDW: true},
	0x2192: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x0, 0x2212, 0x2192}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x2192, HasAlias: true},
	0x2193: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.888}, Stretch: []rune{0x0, 0x23D0, 0x2193}, HDW: [3]float64{0.6, 0, 0.667}, HasHDW: true},
	0x2194: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x2190, 0x2212, 0x2192}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x2194, HasAlias: true},
	0x2195: Delimiter{Direction: DirectionVertical, Sizes: []float64{1.044}, Stretch: []rune{0x2191, 0x23D0, 0x2193}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true},
	0x219E: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x219E, 0x2212}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true},
	0x21A0: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x0, 0x2212, 0x21A0}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true},
	0x21A4: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x2190, 0x2212, 0x2223}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 1.278, HasMin: true, Alias: 0x21A4, HasAlias: true},
	0x21A5: Delimiter{Direction: DirectionVertical, Stretch: []rune{0x2191, 0x23D0, 0x22A5}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true, Min: 1.555, HasMin: true},
	0x21A6: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x2223, 0x2212, 0x2192}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x21A6, HasAlias: true},
	0x21A7: Delimiter{Direction: DirectionVertical, Stretch: []rune{0x22A4, 0x23D0, 0x2193}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true, Min: 1.555, HasMin: true},
	0x21B0: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.722}, Stretch: []rune{0x21B0, 0x23D0}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true},
	0x21B1: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.722}, Stretch: []rune{0x21B1, 0x23D0}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true},
	0x21BC: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x21BC, 0x2212}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true},
	0x21BD: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x21BD, 0x2212}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true},
	0x21BE: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.888}, Stretch: []rune{0x21BE, 0x23D0}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true},
	0x21BF: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.888}, Stretch: []rune{0x21BF, 0x23D0}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true},
	0x21C0: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x0, 0x2212, 0x21C0}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true},
	0x21C1: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x0, 0x2212, 0x21C1}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true},
	0x21C2: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.888}, Stretch: []rune{0x0, 0x23D0, 0x21C2}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true},
	0x21C3: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.888}, Stretch: []rune{0x0, 0x23D0, 0x21C3}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true},
	0x21D0: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x21D0, 0x3D}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x21D0, HasAlias: true},
	0x21D1: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.888}, Stretch: []rune{0x21D1, 0x2016}, HDW: [3]float64{0.599, 0, 0.778}, HasHDW: true},
	0x21D2: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x0, 0x3D, 0x21D2}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x21D2, HasAlias: true},
	0x21D3: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.888}, Stretch: []rune{0x0, 0x2016, 0x21D3}, HDW: [3]float64{0.6, 0, 0.778}, HasHDW: true},
	0x21D4: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x21D0, 0x3D, 0x21D2}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x21D4, HasAlias: true},
	0x21D5: Delimiter{Direction: DirectionVertical, Sizes: []float64{1.044}, Stretch: []rune{0x21D1, 0x2016, 0x21D3}, HDW: [3]float64{0.75, 0.25, 0.778}, HasHDW: true},
	0x21DA: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x21DA, 0x2261}, HDW: [3]float64{0.464, -0.036, 0.5}, HasHDW: true},
	0x21DB: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x0, 0x2261, 0x21DB}, HDW: [3]float64{0.464, -0.036, 0.5}, HasHDW: true},
	0x2212: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.778}, Stretch: []rune{0x0, 0x2212}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x2212, HasAlias: true},
	0x2215: Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Alias: 0x2F, HasAlias: true},
	0x221A: Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Stretch: []rune{0xE001, 0xE000, 0x23B7}, HDW: [3]float64{0.85, 0.35, 1.056}, HasHDW: true, FullExt: [2]float64{0.65, 2.3}, HasFullExt: true},
	0x2223: Delimiter{Direction: DirectionVertical, Sizes: []float64{1}, Stretch: []rune{0x0, 0x2223}, HDW: [3]float64{0.627, 0.015, 0.333}, HasHDW: true, Alias: 0x2223, HasAlias: true},
	0x2225: Delimiter{Direction: DirectionVertical, Sizes: []float64{1}, Stretch: []rune{0x0, 0x2225}, HDW: [3]float64{0.627, 0.015, 0.556}, HasHDW: true},
	0x2308: Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Stretch: []rune{0x23A1, 0x23A2}, HDW: [3]float64{0.85, 0.349, 0.667}, HasHDW: true},
	0x2309: Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Stretch: []rune{0x23A4, 0x23A5}, HDW: [3]float64{0.85, 0.349, 0.667}, HasHDW: true},
	0x230A: Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Stretch: []rune{0x0, 0x23A2, 0x23A3}, HDW: [3]float64{0.85, 0.349, 0.667}, HasHDW: true},
	0x230B: Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Stretch: []rune{0x0, 0x23A5, 0x23A6}, HDW: [3]float64{0.85, 0.349, 0.667}, HasHDW: true},
	0x2312: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.778, 1}, SizeVariants: []int{5, 0}, SizeChars: []rune{0x2322, 0x2322}, Stretch: []rune{0xE150, 0xE154, 0xE151}, HDW: [3]float64{0.32, 0.2, 0.5}, HasHDW: true, Alias: 0x23DC, HasAlias: true},
	0x2322: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.778, 1}, SizeVariants: []int{5, 0}, SizeChars: []rune{0x2322, 0x2322}, Stretch: []rune{0xE150, 0xE154, 0xE151}, HDW: [3]float64{0.32, 0.2, 0.5}, HasHDW: true, Alias: 0x23DC, HasAlias: true},
	0x2323: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.778, 1}, SizeVariants: []int{5, 0}, SizeChars: []rune{0x2323, 0x2323}, Stretch: []rune{0xE152, 0xE154, 0xE153}, HDW: [3]float64{0.32, 0.2, 0.5}, HasHDW: true, Alias: 0x23DD, HasAlias: true},
	0x2329: Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Alias: 0x27E8, HasAlias: true},
	0x232A: Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Alias: 0x27E9, HasAlias: true},
	0x23AA: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.32}, Stretch: []rune{0x23AA, 0x23AA, 0x23AA}, HDW: [3]float64{0.29, 0.015, 0.889}, HasHDW: true},
	0x23AF: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5}, Stretch: []rune{0x0, 0x2013}, HDW: [3]float64{0.285, -0.248, 0.5}, HasHDW: true, Alias: 0x2013, HasAlias: true},
	0x23B0: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.989}, Stretch: []rune{0x23A7, 0x23AA, 0x23AD}, HDW: [3]float64{0.75, 0.25, 0.889}, HasHDW: true},
	0x23B1: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.989}, Stretch: []rune{0x23AB, 0x23AA, 0x23A9}, HDW: [3]float64{0.75, 0.25, 0.889}, HasHDW: true},
	0x23B4: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x250C, 0x2212, 0x2510}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 1, HasMin: true},
	0x23B5: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x2514, 0x2212, 0x2518}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 1, HasMin: true},
	0x23D0: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.602, 1}, SizeVariants: []int{1, 0}, SizeChars: []rune{0x0, 0x2223}, Stretch: []rune{0x0, 0x2223}, HDW: [3]float64{0.602, 0, 0.333}, HasHDW: true},
	0x23DC: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.778, 1}, SizeVariants: []int{5, 0}, SizeChars: []rune{0x2322, 0x2322}, Stretch: []rune{0xE150, 0xE154, 0xE151}, HDW: [3]float64{0.32, 0.2, 0.5}, HasHDW: true, Alias: 0x23DC, HasAlias: true},
	0x23DD: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.778, 1}, SizeVariants: []int{5, 0}, SizeChars: []rune{0x2323, 0x2323}, Stretch: []rune{0xE152, 0xE154, 0xE153}, HDW: [3]float64{0.32, 0.2, 0.5}, HasHDW: true, Alias: 0x23DD, HasAlias: true},
	0x23DE: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0xE150, 0xE154, 0xE151, 0xE155}, HDW: [3]float64{0.32, 0.2, 0.5}, HasHDW: true, Min: 1.8, HasMin: true, Alias: 0x23DE, HasAlias: true},
	0x23DF: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0xE152, 0xE154, 0xE153, 0xE156}, HDW: [3]float64{0.32, 0.2, 0.5}, HasHDW: true, Min: 1.8, HasMin: true, Alias: 0x23DF, HasAlias: true},
	0x23E0: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x2CA, 0x2C9, 0x2CB}, HDW: [3]float64{0.59, -0.544, 0.5}, HasHDW: true, Min: 1, HasMin: true},
	0x23E1: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x2CB, 0x2C9, 0x2CA}, HDW: [3]float64{0.59, -0.544, 0.5}, HasHDW: true, Min: 1, HasMin: true},
	0x2500: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{0.5}, Stretch: []rune{0x0, 0x2013}, HDW: [3]float64{0.285, -0.248, 0.5}, HasHDW: true, Alias: 0x2013, HasAlias: true},
	0x2758: Delimiter{Direction: DirectionVertical, Sizes: []float64{1}, Stretch: []rune{0x0, 0x2223}, HDW: [3]float64{0.627, 0.015, 0.333}, HasHDW: true, Alias: 0x2223, HasAlias: true},
	0x27E8: Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Alias: 0x27E8, HasAlias: true},
	0x27E9: Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Alias: 0x27E9, HasAlias: true},
	0x27EE: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.989}, Stretch: []rune{0x23A7, 0x23AA, 0x23A9}, HDW: [3]float64{0.75, 0.25, 0.889}, HasHDW: true},
	0x27EF: Delimiter{Direction: DirectionVertical, Sizes: []float64{0.989}, Stretch: []rune{0x23AB, 0x23AA, 0x23AD}, HDW: [3]float64{0.75, 0.25, 0.889}, HasHDW: true},
	0x27F5: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x2190, 0x2212}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x2190, HasAlias: true},
	0x27F6: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x0, 0x2212, 0x2192}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x2192, HasAlias: true},
	0x27F7: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x2190, 0x2212, 0x2192}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x2194, HasAlias: true},
	0x27F8: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x21D0, 0x3D}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x21D0, HasAlias: true},
	0x27F9: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x0, 0x3D, 0x21D2}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x21D2, HasAlias: true},
	0x27FA: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x21D0, 0x3D, 0x21D2}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x21D4, HasAlias: true},
	0x27FB: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x2190, 0x2212, 0x2223}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 1.278, HasMin: true, Alias: 0x21A4, HasAlias: true},
	0x27FC: Delimiter{Direction: DirectionHorizontal, Sizes: []float64{1}, Stretch: []rune{0x2223, 0x2212, 0x2192}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Alias: 0x21A6, HasAlias: true},
	0x27FD: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x21D0, 0x3D, 0x2223}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 1.278, HasMin: true, Alias: 0x2906, HasAlias: true},
	0x27FE: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x22A8, 0x3D, 0x21D2}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 1.278, HasMin: true, Alias: 0x2907, HasAlias: true},
	0x2906: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x21D0, 0x3D, 0x2223}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 1.278, HasMin: true, Alias: 0x2906, HasAlias: true},
	0x2907: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x22A8, 0x3D, 0x21D2}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 1.278, HasMin: true, Alias: 0x2907, HasAlias: true},
	0x294E: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x21BC, 0x2212, 0x21C0}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 2, HasMin: true},
	0x294F: Delimiter{Direction: DirectionVertical, Stretch: []rune{0x21BE, 0x23D0, 0x21C2}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true, Min: 1.776, HasMin: true},
	0x2950: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x21BD, 0x2212, 0x21C1}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 2, HasMin: true},
	0x2951: Delimiter{Direction: DirectionVertical, Stretch: []rune{0x21BF, 0x23D0, 0x21C3}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true, Min: 0.5, HasMin: true},
	0x295A: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x21BC, 0x2212, 0x2223}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 1.278, HasMin: true},
	0x295B: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x2223, 0x2212, 0x21C0}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 1.278, HasMin: true},
	0x295C: Delimiter{Direction: DirectionVertical, Stretch: []rune{0x21BE, 0x23D0, 0x22A5}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true, Min: 1.556, HasMin: true},
	0x295D: Delimiter{Direction: DirectionVertical, Stretch: []rune{0x22A4, 0x23D0, 0x21C2}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true, Min: 1.556, HasMin: true},
	0x295E: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x21BD, 0x2212, 0x2223}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 1.278, HasMin: true},
	0x295F: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0x2223, 0x2212, 0x21C1}, HDW: [3]float64{0.583, 0.082, 0.5}, HasHDW: true, Min: 1.278, HasMin: true},
	0x2960: Delimiter{Direction: DirectionVertical, Stretch: []rune{0x21BF, 0x23D0, 0x22A5}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true, Min: 1.776, HasMin: true},
	0x2961: Delimiter{Direction: DirectionVertical, Stretch: []rune{0x22A4, 0x23D0, 0x21C3}, HDW: [3]float64{0.75, 0.25, 0.875}, HasHDW: true, Min: 1.776, HasMin: true},
	0x3008: Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Alias: 0x27E8, HasAlias: true},
	0x3009: Delimiter{Direction: DirectionVertical, Sizes: []float64{1, 1.2, 1.8, 2.4, 3}, Alias: 0x27E9, HasAlias: true},
	0xFE37: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0xE150, 0xE154, 0xE151, 0xE155}, HDW: [3]float64{0.32, 0.2, 0.5}, HasHDW: true, Min: 1.8, HasMin: true, Alias: 0x23DE, HasAlias: true},
	0xFE38: Delimiter{Direction: DirectionHorizontal, Stretch: []rune{0xE152, 0xE154, 0xE153, 0xE156}, HDW: [3]float64{0.32, 0.2, 0.5}, HasHDW: true, Min: 1.8, HasMin: true, Alias: 0x23DF, HasAlias: true},
}
