package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var EvergladeGreen = d2themes.Theme{
	ID:   104,
	Name: "Everglade Green",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.WarmNeutral,

		B1: "#023324",
		B2: "#048E63",
		B3: "#49BC99",
		B4: "#A6E2D0",
		B5: "#CAF2E6",
		B6: "#EBFDF7",

		AA2: "#D35F0A",
		AA4: "#FABA8A",
		AA5: "#FFE0C7",

		AB4: "#C9B9A1",
		AB5: "#E9DBCA",
	},
}
