// generated by go run gen.go; DO NOT EDIT

package sfnt

const numBuiltInPostNames = 258

const builtInPostNamesData = "" +
	".notdef.nullnonmarkingreturnspaceexclamquotedblnumbersigndollarp" +
	"ercentampersandquotesingleparenleftparenrightasteriskpluscommahy" +
	"phenperiodslashzeroonetwothreefourfivesixseveneightninecolonsemi" +
	"colonlessequalgreaterquestionatABCDEFGHIJKLMNOPQRSTUVWXYZbracket" +
	"leftbackslashbracketrightasciicircumunderscoregraveabcdefghijklm" +
	"nopqrstuvwxyzbraceleftbarbracerightasciitildeAdieresisAringCcedi" +
	"llaEacuteNtildeOdieresisUdieresisaacuteagraveacircumflexadieresi" +
	"satildearingccedillaeacuteegraveecircumflexedieresisiacuteigrave" +
	"icircumflexidieresisntildeoacuteograveocircumflexodieresisotilde" +
	"uacuteugraveucircumflexudieresisdaggerdegreecentsterlingsectionb" +
	"ulletparagraphgermandblsregisteredcopyrighttrademarkacutedieresi" +
	"snotequalAEOslashinfinityplusminuslessequalgreaterequalyenmupart" +
	"ialdiffsummationproductpiintegralordfeminineordmasculineOmegaaeo" +
	"slashquestiondownexclamdownlogicalnotradicalflorinapproxequalDel" +
	"taguillemotleftguillemotrightellipsisnonbreakingspaceAgraveAtild" +
	"eOtildeOEoeendashemdashquotedblleftquotedblrightquoteleftquoteri" +
	"ghtdividelozengeydieresisYdieresisfractioncurrencyguilsinglleftg" +
	"uilsinglrightfifldaggerdblperiodcenteredquotesinglbasequotedblba" +
	"seperthousandAcircumflexEcircumflexAacuteEdieresisEgraveIacuteIc" +
	"ircumflexIdieresisIgraveOacuteOcircumflexappleOgraveUacuteUcircu" +
	"mflexUgravedotlessicircumflextildemacronbrevedotaccentringcedill" +
	"ahungarumlautogonekcaronLslashlslashScaronscaronZcaronzcaronbrok" +
	"enbarEthethYacuteyacuteThornthornminusmultiplyonesuperiortwosupe" +
	"riorthreesuperioronehalfonequarterthreequartersfrancGbrevegbreve" +
	"IdotaccentScedillascedillaCacutecacuteCcaronccarondcroat"

var builtInPostNamesOffsets = [...]uint16{
	0x0000, 0x0007, 0x000c, 0x001c, 0x0021, 0x0027, 0x002f, 0x0039,
	0x003f, 0x0046, 0x004f, 0x005a, 0x0063, 0x006d, 0x0075, 0x0079,
	0x007e, 0x0084, 0x008a, 0x008f, 0x0093, 0x0096, 0x0099, 0x009e,
	0x00a2, 0x00a6, 0x00a9, 0x00ae, 0x00b3, 0x00b7, 0x00bc, 0x00c5,
	0x00c9, 0x00ce, 0x00d5, 0x00dd, 0x00df, 0x00e0, 0x00e1, 0x00e2,
	0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7, 0x00e8, 0x00e9, 0x00ea,
	0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef, 0x00f0, 0x00f1, 0x00f2,
	0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00f7, 0x00f8, 0x00f9, 0x0104,
	0x010d, 0x0119, 0x0124, 0x012e, 0x0133, 0x0134, 0x0135, 0x0136,
	0x0137, 0x0138, 0x0139, 0x013a, 0x013b, 0x013c, 0x013d, 0x013e,
	0x013f, 0x0140, 0x0141, 0x0142, 0x0143, 0x0144, 0x0145, 0x0146,
	0x0147, 0x0148, 0x0149, 0x014a, 0x014b, 0x014c, 0x014d, 0x0156,
	0x0159, 0x0163, 0x016d, 0x0176, 0x017b, 0x0183, 0x0189, 0x018f,
	0x0198, 0x01a1, 0x01a7, 0x01ad, 0x01b8, 0x01c1, 0x01c7, 0x01cc,
	0x01d4, 0x01da, 0x01e0, 0x01eb, 0x01f4, 0x01fa, 0x0200, 0x020b,
	0x0214, 0x021a, 0x0220, 0x0226, 0x0231, 0x023a, 0x0240, 0x0246,
	0x024c, 0x0257, 0x0260, 0x0266, 0x026c, 0x0270, 0x0278, 0x027f,
	0x0285, 0x028e, 0x0298, 0x02a2, 0x02ab, 0x02b4, 0x02b9, 0x02c1,
	0x02c9, 0x02cb, 0x02d1, 0x02d9, 0x02e2, 0x02eb, 0x02f7, 0x02fa,
	0x02fc, 0x0307, 0x0310, 0x0317, 0x0319, 0x0321, 0x032c, 0x0338,
	0x033d, 0x033f, 0x0345, 0x0351, 0x035b, 0x0365, 0x036c, 0x0372,
	0x037d, 0x0382, 0x038f, 0x039d, 0x03a5, 0x03b5, 0x03bb, 0x03c1,
	0x03c7, 0x03c9, 0x03cb, 0x03d1, 0x03d7, 0x03e3, 0x03f0, 0x03f9,
	0x0403, 0x0409, 0x0410, 0x0419, 0x0422, 0x042a, 0x0432, 0x043f,
	0x044d, 0x044f, 0x0451, 0x045a, 0x0468, 0x0476, 0x0482, 0x048d,
	0x0498, 0x04a3, 0x04a9, 0x04b2, 0x04b8, 0x04be, 0x04c9, 0x04d2,
	0x04d8, 0x04de, 0x04e9, 0x04ee, 0x04f4, 0x04fa, 0x0505, 0x050b,
	0x0513, 0x051d, 0x0522, 0x0528, 0x052d, 0x0536, 0x053a, 0x0541,
	0x054d, 0x0553, 0x0558, 0x055e, 0x0564, 0x056a, 0x0570, 0x0576,
	0x057c, 0x0585, 0x0588, 0x058b, 0x0591, 0x0597, 0x059c, 0x05a1,
	0x05a6, 0x05ae, 0x05b9, 0x05c4, 0x05d1, 0x05d8, 0x05e2, 0x05ef,
	0x05f4, 0x05fa, 0x0600, 0x060a, 0x0612, 0x061a, 0x0620, 0x0626,
	0x062c, 0x0632, 0x0638,
}
