package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var ShirleyTemple = d2themes.Theme{
	ID:   102,
	Name: "Shirley Temple",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.WarmNeutral,

		B1: "#31021D",
		B2: "#9B1A48",
		B3: "#D2517F",
		B4: "#EA99B6",
		B5: "#FFDAE7",
		B6: "#FCEDF2",

		AA2: "#D35F0A",
		AA4: "#FABA8A",
		AA5: "#FFE0C7",

		AB4: "#FFE767",
		AB5: "#FFF2AA",
	},
}
