package styles

// Present for backwards compatibility.
//
// Deprecated: use styles.Get(name) instead.
var (
	Abap                = Registry["abap"]
	Algol               = Registry["algol"]
	AlgolNu             = Registry["algol_nu"]
	Arduino             = Registry["arduino"]
	Autumn              = Registry["autumn"]
	Average             = Registry["average"]
	Base16Snazzy        = Registry["base16-snazzy"]
	Borland             = Registry["borland"]
	BlackWhite          = Registry["bw"]
	CatppuccinFrappe    = Registry["catppuccin-frappe"]
	CatppuccinLatte     = Registry["catppuccin-latte"]
	CatppuccinMacchiato = Registry["catppuccin-macchiato"]
	CatppuccinMocha     = Registry["catppuccin-mocha"]
	Colorful            = Registry["colorful"]
	DoomOne             = Registry["doom-one"]
	DoomOne2            = Registry["doom-one2"]
	Dracula             = Registry["dracula"]
	Emacs               = Registry["emacs"]
	Friendly            = Registry["friendly"]
	Fruity              = Registry["fruity"]
	GitHubDark          = Registry["github-dark"]
	GitHub              = Registry["github"]
	GruvboxLight        = Registry["gruvbox-light"]
	Gruvbox             = Registry["gruvbox"]
	HrDark              = Registry["hrdark"]
	HrHighContrast      = Registry["hr_high_contrast"]
	Igor                = Registry["igor"]
	KanagawaDragon      = Registry["kanagawa-dragon"]
	KanagawaLotus       = Registry["kanagawa-lotus"]
	KanagawaWave        = Registry["kanagawa-wave"]
	Lovelace            = Registry["lovelace"]
	Manni               = Registry["manni"]
	ModusOperandi       = Registry["modus-operandi"]
	ModusVivendi        = Registry["modus-vivendi"]
	Monokai             = Registry["monokai"]
	MonokaiLight        = Registry["monokailight"]
	Murphy              = Registry["murphy"]
	Native              = Registry["native"]
	Nord                = Registry["nord"]
	OnesEnterprise      = Registry["onesenterprise"]
	ParaisoDark         = Registry["paraiso-dark"]
	ParaisoLight        = Registry["paraiso-light"]
	Pastie              = Registry["pastie"]
	Perldoc             = Registry["perldoc"]
	Pygments            = Registry["pygments"]
	RainbowDash         = Registry["rainbow_dash"]
	RosePineDawn        = Registry["rose-pine-dawn"]
	RosePineMoon        = Registry["rose-pine-moon"]
	RosePine            = Registry["rose-pine"]
	Rrt                 = Registry["rrt"]
	SolarizedDark       = Registry["solarized-dark"]
	SolarizedDark256    = Registry["solarized-dark256"]
	SolarizedLight      = Registry["solarized-light"]
	SwapOff             = Registry["swapoff"]
	Tango               = Registry["tango"]
	Trac                = Registry["trac"]
	Vim                 = Registry["vim"]
	VisualStudio        = Registry["vs"]
	Vulcan              = Registry["vulcan"]
	WitchHazel          = Registry["witchhazel"]
	XcodeDark           = Registry["xcode-dark"]
	Xcode               = Registry["xcode"]
)
