package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var NeutralDefault = d2themes.Theme{
	ID:   0,
	Name: "Neutral Default",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.CoolNeutral,

		B1: "#0D32B2",
		B2: "#0D32B2",
		B3: "#E3E9FD",
		B4: "#E3E9FD",
		B5: "#EDF0FD",
		B6: "#F7F8FE",

		AA2: "#4A6FF3",
		AA4: "#EDF0FD",
		AA5: "#F7F8FE",

		AB4: "#EDF0FD",
		AB5: "#F7F8FE",
	},
}
