// Copyright (c) 2018-2022 The MathJax Consortium
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Derived mechanically from MathJax 3.2.2
// ts/output/{common,svg}/fonts/tex.
// Code generated by generate.py; DO NOT EDIT.

package font

var glyphTables = map[Variant]map[rune]glyphData{
	Normal:              glyphTableNormal,
	Bold:                glyphTableBold,
	Italic:              glyphTableItalic,
	BoldItalic:          glyphTableBoldItalic,
	DoubleStruck:        glyphTableDoubleStruck,
	Fraktur:             glyphTableFraktur,
	BoldFraktur:         glyphTableBoldFraktur,
	Script:              glyphTableScript,
	BoldScript:          glyphTableBoldScript,
	SansSerif:           glyphTableSansSerif,
	BoldSansSerif:       glyphTableBoldSansSerif,
	SansSerifItalic:     glyphTableSansSerifItalic,
	SansSerifBoldItalic: glyphTableSansSerifBoldItalic,
	Monospace:           glyphTableMonospace,
	SmallOp:             glyphTableSmallOp,
	LargeOp:             glyphTableLargeOp,
	Size3:               glyphTableSize3,
	Size4:               glyphTableSize4,
	TeXCalligraphic:     glyphTableTeXCalligraphic,
	TeXBoldCalligraphic: glyphTableTeXBoldCalligraphic,
	TeXMathItalic:       glyphTableTeXMathItalic,
	TeXOldStyle:         glyphTableTeXOldStyle,
	TeXBoldOldStyle:     glyphTableTeXBoldOldStyle,
	TeXVariant:          glyphTableTeXVariant,
}

// Source hashes cover sorted relative names, NUL, exact source bytes, and NUL.
const CommonSourceSHA256 = "6270255b8b5b5c21e09a62261be09d677860135be814d0d031f09b3b4c6e8c28"
const SVGSourceSHA256 = "bfc1e17685cb4d3a30eb3ee674f15c6b95519a745030fc3b4d3126cda98ba888"
const DelimiterSourceSHA256 = "440c69fd81371a8ae766accdd2f0ffe560520d0b1c757d8ed4874a454301e174"
