package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var ButteredToast = d2themes.Theme{
	ID:   105,
	Name: "Buttered Toast",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.WarmNeutral,

		B1: "#312102",
		B2: "#DF9C18",
		B3: "#FDC659",
		B4: "#FFDA90",
		B5: "#FFF0D1",
		B6: "#FFF7E7",

		AA2: "#55452F",
		AA4: "#C9B9A1",
		AA5: "#E9DBCA",

		AB4: "#FABA8A",
		AB5: "#FFE0C7",
	},
}
