package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var Aubergine = d2themes.Theme{
	ID:   7,
	Name: "Aubergine",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.CoolNeutral,

		B1: "#170034",
		B2: "#7639C5",
		B3: "#8F70D1",
		B4: "#D0B9F5",
		B5: "#E7DEFF",
		B6: "#F4F0FF",

		AA2: "#0F66B7",
		AA4: "#87BFF3",
		AA5: "#BCDDFB",

		AB4: "#92E3E3",
		AB5: "#D7F5F5",
	},
}
