package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var MixedBerryBlue = d2themes.Theme{
	ID:   5,
	Name: "Mixed Berry Blue",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.CoolNeutral,

		B1: "#000536",
		B2: "#0F66B7",
		B3: "#4393DD",
		B4: "#87BFF3",
		B5: "#BCDDFB",
		B6: "#E5F3FF",

		AA2: "#7639C5",
		AA4: "#C1A2F3",
		AA5: "#DACEFB",

		AB4: "#EA99C6",
		AB5: "#FFDEF1",
	},
}
