package d2themescatalog

import "github.com/d2lang/d2/d2themes"

var NeutralGrey = d2themes.Theme{
	ID:   1,
	Name: "Neutral Grey",
	Colors: d2themes.ColorPalette{
		Neutrals: d2themes.CoolNeutral,

		B1: "#0A0F25",
		B2: "#676C7E",
		B3: "#9499AB",
		B4: "#CFD2DD",
		B5: "#DEE1EB",
		B6: "#EEF1F8",

		AA2: "#676C7E",
		AA4: "#CFD2DD",
		AA5: "#DEE1EB",

		AB4: "#CFD2DD",
		AB5: "#DEE1EB",
	},
}
